from mpf.core.platform import SwitchConfig
from mpf.core.rgb_color import RGBColor
from mpf.platforms.pkone.pkone import PKONEHardwarePlatform
from mpf.platforms.pkone.pkone_coil import PKONECoilNumber
from mpf.platforms.pkone.pkone_servo import PKONEServoNumber
from mpf.tests.MpfTestCase import MpfTestCase

from mpf.tests.loop import MockSerial


class BaseMockPKONE(MockSerial):

    def __init__(self):
        super().__init__()
        self.queue = []
        self.expected_commands = {}
        self.ignore_commands = {}
        self.sent_commands = []
        self.validate_expected_commands_mode = True

    def reset(self):
        self.expected_commands = {}
        self.sent_commands = []
        self.validate_expected_commands_mode = True

    def read(self, length):
        del length
        if not self.queue:
            return
        msg = (self.queue.pop() + 'E').encode('ascii', 'replace')
        return msg

    def read_ready(self):
        return bool(len(self.queue) > 0)

    def write_ready(self):
        return True

    def _parse(self, msg):
        return False

    def write(self, msg):
        parts = msg.split(b'E')
        # remove last newline
        assert parts.pop() == b''

        for part in parts:
            self._handle_msg(part)

        return len(msg)

    def _handle_msg(self, msg):
        msg_len = len(msg)
        cmd = msg.decode()

        # handle watchdog message
        if cmd == "PWD":
            self.queue.append("PWD")
            return msg_len

        if cmd in self.ignore_commands:
            return msg_len

        if self._parse(cmd):
            return msg_len

        if self.validate_expected_commands_mode:
            if cmd in self.expected_commands:
                if self.expected_commands[cmd]:
                    self.queue.append(self.expected_commands[cmd])
                del self.expected_commands[cmd]
                self.sent_commands.append(cmd)
                return msg_len
            else:
                raise Exception(str(cmd))
        else:
            self.sent_commands.append(cmd)
            return msg_len

    def stop(self):
        pass


class TestPKONE(MpfTestCase):

    """Test the Penny K Pinball PKONE hardware platform."""

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/pkone/'

    def get_platform(self):
        return False

    def _mock_loop(self):
        self.clock.mock_serial("com3", self.controller)

    def tearDown(self):
        self.controller.expected_commands = {
            'PRS': 'PRS',
        }
        super().tearDown()
        self.assertFalse(self.controller.expected_commands)

    def setUp(self):
        self.expected_duration = 2
        self.controller = BaseMockPKONE()

        self.controller.expected_commands = {
            'PCN': 'PCNF11H1',          # Nano controller (firmware 1.1, hardware rev 1)
            'PCB0': 'PCB0XF11H2PY',     # Extension board at ID 0 (firmware 1.1, hardware rev 2, high power on)
            'PCB1': 'PCB1XF11H2PN',     # Extension board at ID 1 (firmware 1.1, hardware rev 2, high power off)
            'PCB2': 'PCB2LF10H1RGB',    # Lightshow board at ID 2 (RGB firmware 1.0, hardware rev 1)
            'PCB3': 'PCB2LF10H1RGBW',   # Lightshow board at ID 3 (RGBW firmware 1.0, hardware rev 1)
            'PCB4': 'PCB4N',
            'PCB5': 'PCB5N',
            'PCB6': 'PCB6N',
            'PCB7': 'PCB7N',
            'PRS': 'PRS',
            'PWS1000': 'PWS',
            'PSA0': 'PSA011000000000000000000000000000000000E',
            'PSA1': 'PSA100110000000000000000000000000000000E',
            'PCC1040000000000': 'PCC',
            'PCC1060000000000': 'PCC',
            'PCC0070000000000': 'PCC',
            'PCC1080000000000': 'PCC',
            'PCC1010000000000': 'PCC',
            'PCC1020000000000': 'PCC',
            'PSC011003': 'PSC',
            'PSC014125': 'PSC',
            'PLB2101050000000000000000000000000000000000000000000000000': 'PLB',
            'PWB3101040000000000000000000000000000000000000000000000000000': 'PWB',
        }

        super().setUp()
        self.advance_time_and_run()
        self.assertFalse(self.controller.expected_commands)

        # test add-on board detection
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        self.assertEqual(2, len(self.machine.default_platform.pkone_extensions))
        self.assertEqual(35, self.machine.default_platform.pkone_extensions[0].switch_count)
        self.assertEqual(10, self.machine.default_platform.pkone_extensions[0].coil_count)
        self.assertEqual(4, self.machine.default_platform.pkone_extensions[0].servo_count)
        self.assertEqual(0, self.machine.default_platform.pkone_extensions[0].addr)

        self.assertEqual(35, self.machine.default_platform.pkone_extensions[1].switch_count)
        self.assertEqual(10, self.machine.default_platform.pkone_extensions[1].coil_count)
        self.assertEqual(4, self.machine.default_platform.pkone_extensions[1].servo_count)
        self.assertEqual(1, self.machine.default_platform.pkone_extensions[1].addr)

        self.assertEqual(2, len(self.machine.default_platform.pkone_lightshows))
        self.assertFalse(self.machine.default_platform.pkone_lightshows[2].rgbw_firmware)
        self.assertEqual(40, self.machine.default_platform.pkone_lightshows[2].simple_led_count)
        self.assertEqual(8, self.machine.default_platform.pkone_lightshows[2].led_groups)
        self.assertEqual(64, self.machine.default_platform.pkone_lightshows[2].max_leds_per_group)
        self.assertEqual(2, self.machine.default_platform.pkone_lightshows[2].addr)

        self.assertTrue(self.machine.default_platform.pkone_lightshows[3].rgbw_firmware)
        self.assertEqual(40, self.machine.default_platform.pkone_lightshows[3].simple_led_count)
        self.assertEqual(8, self.machine.default_platform.pkone_lightshows[3].led_groups)
        self.assertEqual(64, self.machine.default_platform.pkone_lightshows[3].max_leds_per_group)
        self.assertEqual(3, self.machine.default_platform.pkone_lightshows[3].addr)

        self.assertEqual("1.1", self.machine.variables.get_machine_var("pkone_firmware"))
        self.assertEqual("PKONE Nano Controller (rev 1)", self.machine.variables.get_machine_var("pkone_hardware"))

    def test_coils(self):
        self._test_pulse()
        self._test_long_pulse()
        self._test_enable_exception()
        self._test_allow_enable()
        self._test_coil_configure()

        # test hardware scan
        info_str = """Penny K Pinball Hardware
------------------------
 - Connected Controllers:
   -> PKONE Nano - Port: com3 at 115200 baud (firmware v1.1, hardware rev 1).

 - Extension boards:
   -> Address ID: 0 (firmware v1.1, hardware rev 2)
   -> Address ID: 1 (firmware v1.1, hardware rev 2)

 - Lightshow boards:
   -> Address ID: 2 (RGB firmware v1.0, hardware rev 1)
   -> Address ID: 3 (RGBW firmware v1.0, hardware rev 1)
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_coil_configure(self):
        self.assertEqual("PKONE Extension Board 0", self.machine.coils["c_slingshot_test"].hw_driver.get_board_name())
        self.assertEqual("PKONE Extension Board 1", self.machine.coils["c_test"].hw_driver.get_board_name())
        self.assertEqual("PKONE Extension Board 1", self.machine.coils["c_flipper_hold"].hw_driver.get_board_name())
        # last driver on board
        self.controller.expected_commands = {
            "PCC1100000000000": None
        }
        coil = self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '1-10',
                                                              {"recycle_ms": 10})
        self.assertEqual(PKONECoilNumber(1, 10), coil.number)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # board 0 has 10 coils/drivers. configuring driver 17 should not work
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '0-17',
                                                           {"recycle_ms": 10})

        # board 0 has 10 coils/drivers. configuring driver 11 should not work
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '0-11',
                                                           {"recycle_ms": 10})

        # board 0 has 10 coils/drivers. configuring driver 0 should not work (only 1-10)
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '0-0',
                                                           {"recycle_ms": 10})

        # only extension boards 0-1 exist
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '4-1',
                                                           {"recycle_ms": 10})

        # a lightshow board is at address id 2 (no coils are on a lightshow board)
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '2-1',
                                                           {"recycle_ms": 10})

    def _test_pulse(self):
        self.controller.expected_commands = {
            "PCC1040239900027": None,
            "PCP104": None
        }
        # pulse coil 4f
        self.machine.coils["c_test"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _test_long_pulse(self):
        # enable command
        self.controller.expected_commands = {
            "PCC1080009999000": None,
            "PCH108": None
        }
        self.machine.coils["c_long_pulse"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # disable command
        self.controller.expected_commands = {
            "PCR108": None
        }

        self.advance_time_and_run(1)
        # pulse_ms is 2000ms, so after 1s, this should not be sent
        self.assertTrue(self.controller.expected_commands)

        self.advance_time_and_run(1)
        # but after 2s, it should be
        self.assertFalse(self.controller.expected_commands)

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()
            self.advance_time_and_run(.1)

    def _test_allow_enable(self):
        self.controller.expected_commands = {
            "PCC1060239999000": None,
            "PCH106": None
        }
        self.machine.coils["c_test_allow_enable"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def test_rules(self):
        self._test_enable_exception_hw_rule()
        self._test_two_rules_one_switch()
        self._test_hw_rule_pulse()
        self._test_hw_rule_pulse_inverted_switch()
        self._test_hw_rule_same_board()

    def _test_hw_rule_same_board(self):
        # coil and switch are on different boards
        with self.assertRaises(AssertionError):
            self.machine.autofire_coils["ac_different_boards"].enable()
            self.advance_time_and_run(.1)

        # switch and coil on board with address id 1. should work
        self.controller.expected_commands = {
            "PHR10210300000000109900000": None
        }
        self.machine.autofire_coils["ac_board_3"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _test_enable_exception_hw_rule(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = 1.0
            self.machine.flippers["f_test_single"].enable()

        self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = None

    def _test_two_rules_one_switch(self):
        self.controller.expected_commands = {
            "PHR10410700000000239900027": None,
            "PHR10610700000000239900000": None
        }
        self.post_event("ac_same_switch")
        self.hit_and_release_switch("s_flipper")
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _test_hw_rule_pulse(self):
        self.controller.expected_commands = {
            "PHR00712200000000109900000": None     # hw rule
        }
        self.machine.autofire_coils["ac_slingshot_test"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        self.controller.expected_commands = {
            "PHD007": None
        }
        self.machine.autofire_coils["ac_slingshot_test"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _test_hw_rule_pulse_inverted_switch(self):
        self.controller.expected_commands = {
            "PHR00712610000000109900000": None
        }
        self.machine.autofire_coils["ac_inverted_switch"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def test_servo(self):
        # test servo numbers
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        servo_number = self.machine.default_platform._parse_servo_number("0-11")
        self.assertEqual(11, servo_number.servo_number)
        servo_number = self.machine.default_platform._parse_servo_number("0-14")
        self.assertEqual(14, servo_number.servo_number)

        with self.assertRaises(AssertionError) as e:
            self.machine.default_platform._parse_servo_number("0-1")
            self.machine.default_platform._parse_servo_number("0-9")
            self.machine.default_platform._parse_servo_number("0-15")

        # go to min position
        self.controller.expected_commands = {
                "PSC011003": None
        }
        self.machine.servos["servo1"].go_to_position(0)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # go to max position
        self.controller.expected_commands = {
                "PSC011027": None
        }
        self.machine.servos["servo1"].go_to_position(1)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _switch_hit_cb(self, **kwargs):
        self.switch_hit = True

    def test_switches(self):
        # test initial switch states (based on PSA command response)
        self.assertSwitchState("s_test_1", 1)
        self.assertSwitchState("s_test_2", 1)
        self.assertSwitchState("s_test_3", 0)
        self.assertSwitchState("s_test_4", 0)
        self.assertSwitchState("s_test_11", 0)
        self.assertSwitchState("s_test_12", 0)
        self.assertSwitchState("s_test_13", 1)
        self.assertSwitchState("s_test_14", 1)

        self._test_switch_changes()
        self._test_switch_changes_nc()
        self._test_switch_configure()

    def _test_switch_configure(self):
        # configuring switches in PKONE does not generate any commands between the host and controller

        # last switch on first board
        self.machine.default_platform.configure_switch('0-35', SwitchConfig(name="", debounce='auto', invert=0), {})
        self.advance_time_and_run(.1)

        # next should not work
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('0-36', SwitchConfig(name="", debounce='auto', invert=0), {})

        # invalid board
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('4-1', SwitchConfig(name="", debounce='auto', invert=0), {})

        # invalid switch (switch number begin at 1 and not 0)
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('0-0', SwitchConfig(name="", debounce='auto', invert=0), {})

    def _test_switch_changes(self):
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test", 0)
        self.assertFalse(self.switch_hit)

        self.machine.events.add_handler("s_test_active", self._switch_hit_cb)
        self.machine.default_platform.process_received_message("PSW0071E")
        self.advance_time_and_run(1)

        self.assertTrue(self.switch_hit)
        self.assertSwitchState("s_test", 1)
        self.switch_hit = False

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 1)

        self.machine.default_platform.process_received_message("PSW0070E")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 0)

    def _test_switch_changes_nc(self):
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test_nc", 1)
        self.assertFalse(self.switch_hit)

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 1)

        self.machine.default_platform.process_received_message("PSW0261E")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 0)

        self.machine.events.add_handler("s_test_nc_active", self._switch_hit_cb)
        self.machine.default_platform.process_received_message("PSW0260E")
        self.advance_time_and_run(1)

        self.assertSwitchState("s_test_nc", 1)
        self.assertTrue(self.switch_hit)
        self.switch_hit = False

    def test_flipper_single_coil(self):
        # manual flip no hw rule
        self.controller.expected_commands = {
            "PCC1010109900000": None,
            "PCP101": None,
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual enable no hw rule
        self.controller.expected_commands = {
            "PCH101": None,
            "PCC1010109912000": None
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual disable no hw rule
        self.controller.expected_commands = {
            "PCR101": None
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # flipper rule enable
        self.controller.expected_commands = {
            "PHR10140500000000119912000": None
        }
        self.machine.flippers["f_test_single"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual flip with hw rule in action
        self.controller.expected_commands = {
            "PCP101": None  # pulse (no need to configure because config did not change)
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual flip with hw rule in action without reconfigure (same pulse)
        self.controller.expected_commands = {
            "PCC1010119912000": None,       # reconfigure coil
            "PCP101": None,                 # pulse
        }
        self.machine.coils["c_flipper_main"].pulse(11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual enable with hw rule (same pulse)
        self.controller.expected_commands = {
            "PCH101": None      # pulse (config did not change so no config msg necessary)
        }
        self.machine.coils["c_flipper_main"].enable(pulse_ms=11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual disable with hw rule
        self.controller.expected_commands = {
            "PCR101": None,
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual enable with hw rule (different pulse)
        self.controller.expected_commands = {
            "PCC1010109912000": None,       # configure + enable
            "PCH101": None
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual disable with hw rule
        self.controller.expected_commands = {
            "PCR101": None
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # disable rule
        self.controller.expected_commands = {
            "PHD101": None
        }
        self.machine.flippers["f_test_single"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual flip no hw rule (config is still cached)
        self.controller.expected_commands = {
            "PCP101": None
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # manual flip again with cached config
        self.controller.expected_commands = {
            "PCP101": None
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def test_flipper_two_coils(self):
        # we pulse the main coil
        # hold coil is pulsed + enabled
        self.controller.expected_commands = {
            "PHR10130500000000109900000": None,
            "PHR10240500000000109912000": None,
        }
        self.machine.flippers["f_test_hold"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        self.controller.expected_commands = {
            "PHD101": None,
            "PHD102": None
        }
        self.machine.flippers["f_test_hold"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def test_leds(self):
        self._test_simple_led()
        self._test_led_hardware_alignment()
        self._test_ws281x_led()

    def _test_simple_led(self):
        self.assertTrue("test_simple_led" in self.machine.lights)
        self.assertTrue("test_other_simple_led" in self.machine.lights)

        # test enable of simple led
        self.controller.expected_commands = {
            "PLS2171": None,
            "PLS3011": None,
        }
        self.machine.lights["test_simple_led"].on()
        self.machine.lights["test_other_simple_led"].on()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # test disable of simple led
        self.controller.expected_commands = {
            "PLS2170": None,
            "PLS3010": None,
        }
        self.machine.lights["test_simple_led"].off()
        self.machine.lights["test_other_simple_led"].off()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # test turning on of simple led using color
        self.controller.expected_commands = {
            "PLS2171": None,
            "PLS3011": None,
        }
        self.machine.lights["test_simple_led"].color(RGBColor("FFFFFF"))
        self.machine.lights["test_other_simple_led"].color(RGBColor("FFFFFF"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

        # test turning off of simple led using color
        self.controller.expected_commands = {
            "PLS2170": None,
            "PLS3010": None,
        }
        self.machine.lights["test_simple_led"].color(RGBColor("000000"))
        self.machine.lights["test_other_simple_led"].color(RGBColor("000000"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)

    def _test_ws281x_led(self):
        self.advance_time_and_run()

        # get light devices
        device_rgb_1 = self.machine.lights["test_rgb_led_1"]
        device_rgb_2 = self.machine.lights["test_rgb_led_2"]
        device_rgb_3 = self.machine.lights["test_rgb_led_3"]
        device_rgb_4 = self.machine.lights["test_rgb_led_4"]

        device_rgbw_1 = self.machine.lights["test_rgbw_led_1"]
        device_rgbw_2 = self.machine.lights["test_rgbw_led_2"]
        device_rgbw_3 = self.machine.lights["test_rgbw_led_3"]
        device_rgbw_4 = self.machine.lights["test_rgbw_led_4"]

        # ensure channel numbers (hardware drivers) have been assigned correctly
        self.assertListEqual(["2-1-0", "2-1-1", "2-1-2"], device_rgb_1.get_hw_numbers())
        self.assertListEqual(["2-1-3", "2-1-4", "2-1-5"], device_rgb_2.get_hw_numbers())
        self.assertListEqual(["2-1-6", "2-1-7", "2-1-8", "2-1-9"], device_rgb_3.get_hw_numbers())
        self.assertListEqual(["2-1-10", "2-1-11", "2-1-12"], device_rgb_4.get_hw_numbers())

        self.assertListEqual(["3-1-0", "3-1-1", "3-1-2", "3-1-3"], device_rgbw_1.get_hw_numbers())
        self.assertListEqual(["3-1-4", "3-1-5", "3-1-6", "3-1-7"], device_rgbw_2.get_hw_numbers())
        self.assertListEqual(["3-1-8", "3-1-9", "3-1-10"], device_rgbw_3.get_hw_numbers())
        self.assertListEqual(["3-1-11", "3-1-12", "3-1-13", "3-1-14"], device_rgbw_4.get_hw_numbers())

        # ensure channels were properly added to the Lightshow boards
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        rgb_hw_channels = self.machine.default_platform.pkone_lightshows[2].get_all_channel_hw_drivers()
        rgbw_hw_channels = self.machine.default_platform.pkone_lightshows[3].get_all_channel_hw_drivers()
        self.assertEqual(13, len(rgb_hw_channels))
        self.assertEqual(15, len(rgbw_hw_channels))

        # ensure all LEDs are initially off
        self.assertLightColor("test_rgb_led_1", RGBColor("off"))
        self.assertLightColor("test_rgb_led_2", RGBColor("off"))
        self.assertLightColor("test_rgb_led_3", RGBColor("off"))
        self.assertLightColor("test_rgb_led_4", RGBColor("off"))

        self.assertLightColor("test_rgbw_led_1", RGBColor("off"))
        self.assertLightColor("test_rgbw_led_2", RGBColor("off"))
        self.assertLightColor("test_rgbw_led_3", RGBColor("off"))
        self.assertLightColor("test_rgbw_led_4", RGBColor("off"))

        # test turning on rgb led using color
        self.controller.expected_commands = {
            "PLB2101020000255000000000000255": None,
        }
        self.machine.lights["test_rgb_led_1"].color(RGBColor("red"))
        self.machine.lights["test_rgb_led_2"].color(RGBColor("blue"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgb led using color with fade
        self.controller.expected_commands = {
            "PLB2101020012000128000255000000": None,
        }
        self.machine.lights["test_rgb_led_1"].color(RGBColor("green"), 123)
        self.machine.lights["test_rgb_led_2"].color(RGBColor("red"), 123)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgb led out of hardware alignment using color
        self.controller.expected_commands = {
            "PLB2103020000063000011192000000": None,
        }
        self.machine.lights["test_rgb_led_3"].color(RGBColor("pink"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning off rgb led
        self.controller.expected_commands = {
            "PLB2101050000000000000000000000000000000000000000000000000": None,
        }
        self.machine.lights["test_rgb_led_1"].off()
        self.machine.lights["test_rgb_led_2"].off()
        self.machine.lights["test_rgb_led_3"].off()
        self.machine.lights["test_rgb_led_4"].off()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgb led out of hardware alignment using color with fade (uses software fade)
        self.controller.validate_expected_commands_mode = False
        self.machine.lights["test_rgb_led_3"].color(RGBColor("pink"), 250)
        self.advance_time_and_run(.5)
        self.assertEqual(14, len(self.controller.sent_commands))
        self.assertEqual("PLB2103020000000000000000000000", self.controller.sent_commands[0])
        self.assertEqual("PLB2103020000063000011192000000", self.controller.sent_commands[-1])
        self.controller.reset()

        # test turning off rgb led out of hardware alignment with fade (uses software fade) and turning on
        # an rgb led in alignment with color simultaneously (uses hardware fade)
        self.controller.validate_expected_commands_mode = False
        self.machine.lights["test_rgb_led_1"].color(RGBColor("blue"), 250)
        self.machine.lights["test_rgb_led_3"].off(250)
        self.advance_time_and_run(.5)
        self.assertEqual(15, len(self.controller.sent_commands))
        self.assertEqual("PLB2101010025000000255", self.controller.sent_commands[0])
        self.assertEqual("PLB2103020000063000011192000000", self.controller.sent_commands[1])
        self.assertEqual("PLB2103020000000000000000000000", self.controller.sent_commands[-1])
        self.controller.reset()

        # test turning on rgbw led using color
        self.controller.expected_commands = {
            "PWB3101020000255000000000000000255000": None,
        }
        self.machine.lights["test_rgbw_led_1"].color(RGBColor("red"))
        self.machine.lights["test_rgbw_led_2"].color(RGBColor("blue"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgbw led using color with fade
        self.controller.expected_commands = {
            "PWB3101020012000128000000255000000000": None,
        }
        self.machine.lights["test_rgbw_led_1"].color(RGBColor("green"), 128)
        self.machine.lights["test_rgbw_led_2"].color(RGBColor("red"), 128)
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgbw led out of hardware alignment using color
        self.controller.expected_commands = {
            "PWB3103010000255192203000": None,
        }
        self.machine.lights["test_rgbw_led_3"].color(RGBColor("pink"))
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning off rgbw led
        self.controller.expected_commands = {
            "PWB3101040000000000000000000000000000000000000000000000000000": None,
        }
        self.machine.lights["test_rgbw_led_1"].off()
        self.machine.lights["test_rgbw_led_2"].off()
        self.machine.lights["test_rgbw_led_3"].off()
        self.machine.lights["test_rgbw_led_4"].off()
        self.advance_time_and_run(.1)
        self.assertFalse(self.controller.expected_commands)
        self.controller.reset()

        # test turning on rgbw led out of hardware alignment using color with fade (uses software fade)
        self.controller.validate_expected_commands_mode = False
        self.machine.lights["test_rgbw_led_3"].color(RGBColor("pink"), 250)
        self.advance_time_and_run(.5)
        self.assertEqual(14, len(self.controller.sent_commands))
        self.assertEqual("PWB3103010000000000000000", self.controller.sent_commands[0])
        self.assertEqual("PWB3103010000255192203000", self.controller.sent_commands[-1])
        self.controller.reset()

        # test turning off rgbw led out of hardware alignment with fade (uses software fade) and turning on
        # an rgbw led in alignment with color simultaneously (uses hardware fade)
        self.controller.validate_expected_commands_mode = False
        self.machine.lights["test_rgbw_led_1"].color(RGBColor("blue"), 250)
        self.machine.lights["test_rgbw_led_3"].off(250)
        self.advance_time_and_run(.5)
        self.assertEqual(15, len(self.controller.sent_commands))
        self.assertEqual("PWB3101010025000000255000", self.controller.sent_commands[0])
        self.assertEqual("PWB3103010000255192203000", self.controller.sent_commands[1])
        self.assertEqual("PWB3103010000000000000000", self.controller.sent_commands[-1])
        self.controller.reset()

    def _test_led_hardware_alignment(self):
        self.assertIsInstance(self.machine.default_platform, PKONEHardwarePlatform)
        self.assertTrue(self.machine.default_platform._led_is_hardware_aligned("test_rgb_led_1"))
        self.assertTrue(self.machine.default_platform._led_is_hardware_aligned("test_rgb_led_2"))
        self.assertFalse(self.machine.default_platform._led_is_hardware_aligned("test_rgb_led_3"))
        self.assertFalse(self.machine.default_platform._led_is_hardware_aligned("test_rgb_led_4"))
        self.assertTrue(self.machine.default_platform._led_is_hardware_aligned("test_rgbw_led_1"))
        self.assertTrue(self.machine.default_platform._led_is_hardware_aligned("test_rgbw_led_2"))
        self.assertFalse(self.machine.default_platform._led_is_hardware_aligned("test_rgbw_led_3"))
        self.assertFalse(self.machine.default_platform._led_is_hardware_aligned("test_rgbw_led_4"))



"""Scan serial connections and request ID: to identify FAST hardware."""
import asyncio

from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE, SerialException
from serial.tools import list_ports

from mpf.platforms.fast.fast_defines import USB_IDS


class FastPortDetector:

    """Class for detecting FAST platform hardware on connected serial ports."""

    def __init__(self, platform, autodetect_processors, hardcoded_ports):
        """Initialize the port detector."""
        self.platform = platform
        self.machine = platform.machine
        self.autodetect_processors = autodetect_processors
        self.hardcoded_ports = hardcoded_ports
        self.detected_fast_ports = list()  # tuples of (port, baud)
        self.tasks = list()
        self.task_writers = dict()
        self.results = dict()  # dict of processor: port

        self.platform.log.info("Auto-detecting ports for the following connections: %s", self.autodetect_processors)

        self._find_fast_devices()

    def _find_fast_devices(self):
        # Creates a list of all serial ports which have devices attached which match
        # USB VID/PID combinations of FAST devices we're looking for.
        for port in list_ports.comports():
            if (port.vid, port.pid) in USB_IDS:
                proc, desc = USB_IDS[(port.vid, port.pid)]
                if proc in self.autodetect_processors:
                    baud = self.machine.config['fast'][proc]['baud']
                    if port.device not in self.hardcoded_ports:
                        self.detected_fast_ports.append((port.device, baud))
                        self.platform.log.debug("Port %s is connected to a %s.", port.device, desc)
                    else:
                        self.platform.log.debug(
                            "Skipping auto-detect of %s on %s since it's in the config file elsewhere.",
                            proc, port.device)

    async def detect_ports(self):
        """Query the available serial ports for an ID: response."""
        self.tasks = [asyncio.create_task(self._connect_task(port, baud)) for port, baud in self.detected_fast_ports]
        for task, (_, _) in zip(self.tasks, self.detected_fast_ports):
            task.add_done_callback(self._cleanup_writer)
        # need to catch the task cancellation exception here or else the MPF startup process will think something broke
        await asyncio.gather(*self.tasks, return_exceptions=True)
        await asyncio.sleep(0.1)

    async def _connect_task(self, port, baud):
        """One instance per port is launched to connect and ID whatever's on the other end.

        Sends an ID: command once per second until it gets a response or until other tasks have
        found all the ports and the remaining ones are cancelled.

        If an ID: response comes back that matches a processor we're looking for,
        it reports success and end.

        """
        while True:
            try:
                connector = self.machine.clock.open_serial_connection(
                    url=port, baudrate=baud, limit=0, xonxoff=False,
                    bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE)
                reader, writer = await connector
            except SerialException:
                self.log.error("Could not connect to port %s. Are you connected via CoolTerm? :)", port)

            self.platform.debug_log(" - success connecting to port %s", port)
            self.task_writers[asyncio.current_task()] = writer

            total_attempts = 5
            attempts = 0
            while attempts < total_attempts:
                writer.write(b'ID:\r')

                # Wait for a response with 1-second timeout
                try:
                    data = await asyncio.wait_for(reader.read(100), timeout=1.0)
                except asyncio.TimeoutError:
                    pass

                if data:
                    data = data.decode('utf-8', errors='ignore')

                    for processor in self.autodetect_processors:
                        if processor.upper() in data:
                            self._report_success(processor, port)
                            writer.close()
                            return
                attempts += 1
            self.platform.debug_log("Unable to get ID: on port %s after %s retries.", port, total_attempts)
            return

    def _report_success(self, processor, port):
        self.results[processor] = port
        self.platform.log.info('Detected %s on port %s', processor.upper(), port)
        self.platform.config[processor]['port'] = [port]

        if len(self.results) >= len(self.autodetect_processors):
            self._cancel_remaining_tasks()

    def _cancel_remaining_tasks(self):
        for task in self.tasks:
            if not task.done():
                task.cancel()

    def _cleanup_writer(self, task):
        writer = self.task_writers.pop(task, None)
        if writer:
            writer.close()



"""FAST Audio Board Serial Communicator."""
# mpf/platforms/fast/communicators/aud.py

from packaging import version

from mpf.platforms.fast.communicators.base import FastSerialCommunicator
from mpf.platforms.fast.fast_audio import FASTAudioInterface
from mpf.core.utility_functions import Util


class FastAudCommunicator(FastSerialCommunicator):

    """Handles the serial communication to the FAST platform."""

    # Notes on this class, versus the FASTAudioInterface class in mpf/platforms/fast/fast_audio.py:

    # FastAudCommunicator focuses on actually talking to the hardware. It will deal in hardware
    # levels 0-63, and handle decoding/encoding the config bitmask, etc.
    # FastAudioInterface handles the higher level stuff, like volume scaling, and the general
    # interface into the rest of MPF (reading configs, interacting with events, etc.)

    MIN_FW = version.parse('0.10')
    IGNORED_MESSAGES = ['AV:', 'AS:', 'AH:', 'AM:']

    __slots__ = ["amps", "current_config_byte", "phones_level", "phones_mute", "_watchdog_ms"]

    def __init__(self, platform, processor, config):
        """Initialize AUD communicator."""
        super().__init__(platform, processor, config)
        self.amps = {
            'main':
                {'cmd': 'AV',
                 'volume': 0,
                 'enabled': False,
                 },
            'sub':
                {'cmd': 'AS',
                 'volume': 0,
                 'enabled': False,
                 },
            'headphones':
                {'cmd': 'AH',
                 'volume': 0,
                 'enabled': False,
                 },
        }
        self.current_config_byte = '00'
        self.phones_level = True  # False = line, True = phones
        self.phones_mute = False  # False = ignore phones insertion, True = mute main/sub when phones inserted
        self._watchdog_ms = 0

    async def init(self):
        """Initialize the serial connection and identify the board."""
        # Loop here until we get a response
        await self.send_and_wait_for_response_processed('ID:', 'ID:', max_retries=-1)
        self.platform.audio_interface = FASTAudioInterface(self.platform, self)

    async def soft_reset(self):
        """Send a soft reset command to the board."""
        self.update_config(send_now=True)
        for amp_name, amp in self.amps.items():
            self.set_volume(amp_name, amp['volume'])

    def _volume_to_hw(self, volume):
        """Always store and pass volume levels as decimals (0-64).

        Use this method to convert to hex strings when sending via serial to AUD board.
        """
        volume = int(volume)
        assert 0 <= volume <= 63, f"Invalid volume {volume}"
        return f"{volume:02X}"

    def update_config(self, send_now=True):
        """Write a new configuration to the platform."""
        byte = '00'
        if self.amps['main']['enabled']:
            byte = Util.set_bit(byte, 0)
        if self.amps['sub']['enabled']:
            byte = Util.set_bit(byte, 1)
        if self.amps['headphones']['enabled']:
            if not self.phones_level:  # line level
                byte = Util.set_bit(byte, 2)
                byte = Util.set_bit(byte, 3)
            else:  # phones level
                if self.phones_mute:
                    byte = Util.set_bit(byte, 3)
                else:
                    byte = Util.set_bit(byte, 2)

        self.current_config_byte = byte

        if send_now:
            self.send_config_to_board()

    def set_volume(self, amp_name, volume, send_now=True):
        """Set the volume for a specific amp."""
        if amp_name not in self.amps:
            raise AssertionError(f"Invalid amp {amp_name}")

        # Track the internal (realtime) volume of the amp, which may be
        # different than the stored (machine var) volume during ducking
        self.amps[amp_name]['volume'] = volume
        if send_now:
            hw_volume = self._volume_to_hw(volume)
            self.send_and_forget(f"{self.amps[amp_name]['cmd']}:{hw_volume}")

    def get_volume(self, amp_name):
        """Get the current internal volume of the amp."""
        return self.amps[amp_name]['volume']

    def enable_amp(self, amp_name, send_now=True):
        """Enable a specific amp."""
        if amp_name not in self.amps:
            raise AssertionError(f"Invalid amp {amp_name}")
        self.amps[amp_name]['enabled'] = True
        self.update_config(send_now)

    def disable_amp(self, amp_name, send_now=True):
        """Disable a specific amp."""
        if amp_name not in self.amps:
            raise AssertionError(f"Invalid amp {amp_name}")
        self.amps[amp_name]['enabled'] = False
        self.update_config(send_now)

    def set_phones_level(self, mode, send_now=True):
        """Set the volume level for headphones."""
        if mode == 'line':
            self.phones_level = False
        elif mode == 'headphones':
            self.phones_level = True
        else:
            raise AssertionError(f"Invalid phones level {mode}")

        self.update_config(send_now)

    def set_phones_behavior(self, behavior, send_now=True):
        """Set the audio board behavior when headphones are plugged in."""
        # Will return False if not possible, true if ok
        if not self.phones_level:  # phones are line level, this does not apply
            return False

        if behavior == 'mute':
            self.phones_mute = True
        elif behavior == 'ignore':
            self.phones_mute = False
        else:
            raise AssertionError(f"Invalid phones behavior {behavior}")

        self.update_config(send_now)
        return True

    def send_config_to_board(self):
        """Write the current configuration to the board."""
        self.send_and_forget(f"AM:{self.current_config_byte}")

    def save_settings_to_firmware(self):
        """Save the current configuration on the board's persistent memory."""
        self.send_and_forget("AW:")

    def set_watchdog(self, timeout):
        """Set a watchdog timer on the audio board."""
        # TODO setup task to send watchdog

    def get_phones_status(self):
        """Request the current status of the headphone port."""
        # TODO WD command processing

    def get_power_status(self):
        """Request the current status of the power detection port."""
        # TODO WD command processing

    def pulse_control_pin(self, pin, ms):
        """Pulse an LCD control pin."""
        assert 0 <= pin <= 7, f"Invalid pin {pin}"
        self.send_and_forget(f"XO:{pin:02X},{hex(ms)[2:].upper()}")



"""FAST I/O Board."""

from mpf.platforms.fast.fast_defines import VALID_IO_BOARDS


class FastIoBoard:

    """A FAST I/O board on the NET processor."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, communicator, name, node_id, model_string, firmware_version,
                 switch_count, driver_count, prior_switches, prior_drivers):
        """Initialize FastIoBoard."""
        self.communicator = communicator
        self.name = str(name)
        self.node_id = node_id  # position in the I/O loop, 0-indexed
        self.model = model_string
        self.firmware_version = firmware_version
        self.start_switch = prior_switches
        self.start_driver = prior_drivers
        self.switch_count = switch_count
        self.driver_count = driver_count

        assert self.model in VALID_IO_BOARDS, "Invalid I/O board model: {}".format(self.model)

    def __repr__(self):
        """Return representation of this I/O board."""
        return f'{self.model} "{self.name}"'

    def get_description_string(self) -> str:
        """Return description string."""
        return f"Board {self.node_id} - Model: {self.model}, Firmware: {self.firmware_version}, " + \
            f"Switches: {self.switch_count}, Drivers: {self.driver_count}"



"""A driver/coil in the FAST platform."""
import logging
from copy import copy
from dataclasses import dataclass

from mpf.core.platform import DriverConfig
from mpf.core.utility_functions import Util
from mpf.exceptions.config_file_error import ConfigFileError
from mpf.platforms.fast.communicators.base import FastSerialCommunicator
from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.fast.fast import \
        FastHardwarePlatform  # pylint: disable-msg=cyclic-import,unused-import


@dataclass
class FastDriverConfig:

    """Class for FAST hardware driver configuration."""

    number: str
    trigger: str
    switch_id: str
    mode: str
    param1: str
    param2: str
    param3: str
    param4: str
    param5: str


class FASTDriver:

    """Base class for drivers connected to a FAST Controller."""

    __slots__ = ["log", "communicator", "number", "hw_number", "autofire_config", "baseline_driver_config",
                 "config", "current_driver_config", "mode_param_mapping", "platform_settings"]

    def __init__(self, communicator: FastSerialCommunicator, hw_number: int) -> None:
        """Initialize the driver object.

        This is called once for each physical driver on the connected hardware,
        regardless of whether it's configured in MPF.
        """
        self.log = logging.getLogger('FAST Driver')
        self.communicator = communicator
        self.number = hw_number  # must be int to work with the rest of MPF
        self.hw_number = Util.int_to_hex_string(hw_number)  # hex version the FAST hw actually uses
        self.config = None
        self.autofire_config = None
        self.platform_settings = dict()

        self.baseline_driver_config = FastDriverConfig(number=self.hw_number, trigger='00',
                                                       switch_id='00', mode='00',
                                                       param1='00', param2='00', param3='00',
                                                       param4='00', param5='00')
        self.current_driver_config = self.baseline_driver_config

        self.mode_param_mapping = {
            '10': ['pwm1_ms', 'pwm1_power', 'pwm2_ms', 'pwm2_power', 'recycle_ms'],
            '12': ['pwm1_ms', 'pwm1_power', 'pwm2_ms', 'pwm2_power', 'kick_ms'],
            '18': ['pwm1_ms', 'pwm1_power', 'pwm2_power', 'recycle_ms', None],
            '20': ['off_switch', 'pwm1_ms', 'pwm1_power', 'pwm2_power', 'rest_ms'],
            '25': ['relay_on_report_ms', 'relay_off_report_ms'],
            '30': ['delay_ms_x10', 'pwm1_ms', 'pwm2_ms', 'pwm2_power', 'recycle_ms'],
            '70': ['pwm1_ms', 'pwm1_power', 'pwm2_ms_x100', 'pwm2_power', 'recycle_ms'],
            '75': ['off_switch', 'pwm1_ms', 'pwm2_ms_x100', 'pwm2_power', 'recycle_ms'],
        }

    def set_initial_config(self, mpf_config: DriverConfig, platform_settings):
        """Sets the initial config for this driver based on the MPF config.

        Parameters
        ----------
            mpf_config: DriverConfig instance which holds the MPF DriverConfig settings
                for this driver from the config file. This already incorporates
                any machine-wide defaults, etc. so it's ready to go.
            platform_settings: FastDriverConfig instance which holds any platform_settings:
                entries for this driver from the config file.

        This method does not actually write the config to the driver. Is just figures out
            what the FastDriverConfig should be.

        This will not be called for drivers that are not in the MPF config.
        """
        self.config = mpf_config
        self.platform_settings = platform_settings
        self.current_driver_config = self.convert_mpf_config_to_fast(mpf_config, platform_settings)
        self.baseline_driver_config = copy(self.current_driver_config)

    def convert_mpf_config_to_fast(self, mpf_config: DriverConfig, platform_settings) -> FastDriverConfig:
        """Convert a DriverConfig (used throughout MPF) to FastDriverConfig (FAST specific version).

        This is only used for the initial configuration of drivers. Autofire rules update these.
        """
        if mpf_config.default_recycle:
            raise ConfigFileError("FAST platform does not support default_recycle for coils. "
                                  f"Use platform_settings:recycle_ms instead. Coil '{mpf_config.name}'.",
                                  7, self.log.name)

        if mpf_config.default_pulse_ms > 255:
            raise ConfigFileError("FAST platform does not support default_pulse_ms > 255. Use "
                                  f"platform_settings:pwm2_ms which goes up to 25,500ms. Coil '{mpf_config.name}'.",
                                  7, self.log.name)

        pwm2_ms, pwm2_power, recycle_ms = self._get_platform_settings(mpf_config, platform_settings)

        if platform_settings['pwm2_ms'] and platform_settings['pwm2_ms'] > 255:
            mode = '70'
            pwm2_ms = platform_settings['pwm2_ms'] // 100
        else:
            mode = '10'

        return FastDriverConfig(number=self.hw_number,
                                trigger='81',
                                switch_id='00',
                                mode=mode,
                                param1=Util.int_to_hex_string(mpf_config.default_pulse_ms),
                                param2=Util.float_to_pwm8_hex_string(mpf_config.default_pulse_power),
                                param3=Util.int_to_hex_string(pwm2_ms),
                                param4=Util.float_to_pwm8_hex_string(pwm2_power),
                                param5=Util.int_to_hex_string(recycle_ms))

    def _get_platform_settings(self, mpf_config: DriverConfig, platform_settings):
        if platform_settings['pwm2_ms'] is not None:
            pwm2_ms = platform_settings['pwm2_ms']
        else:
            pwm2_ms = 0

        if mpf_config.default_hold_power is not None:
            pwm2_power = mpf_config.default_hold_power
        else:
            pwm2_power = 0

        if platform_settings['recycle_ms'] is not None:
            recycle_ms = platform_settings['recycle_ms']
        else:
            # mpf_config.default_recycle is a bool and not well defined, so we ignore it
            # in the FAST platform
            recycle_ms = 0

        return pwm2_ms, pwm2_power, recycle_ms

    def set_bit(self, hex_string, bit):
        """Sets a bit in a hex string.

        Parameters
        ----------
            hex_string (_type_): Hex string, e.g. '81'
            bit (_type_): Bit to set, e.g. 3

        Returns
        -------
            _type_: Returns the hex string with the bit set, e.g. '89'
        """
        num = int(hex_string, 16)
        num |= 1 << bit
        return Util.int_to_hex_string(num)

    def clear_bit(self, hex_string, bit):
        """Remove a given bit from a hex string."""
        num = int(hex_string, 16)
        num &= ~(1 << bit)
        return Util.int_to_hex_string(num)

    def send_config_to_driver(self, one_shot: bool = False, wait_to_confirm: bool = False):
        """Send a driver configuration to the platform."""
        self.log.debug("Sending config to driver %s (0x%s). one_shot: %s. wait_to_confirm: %s",
                       self.number, self.hw_number, one_shot, wait_to_confirm)

        if one_shot:
            trigger = self.set_bit(self.current_driver_config.trigger, 3)
        else:
            trigger = self.current_driver_config.trigger

        msg = (f'{self.communicator.DRIVER_CMD}:{self.hw_number},{trigger},'
               f'{self.current_driver_config.switch_id},'
               f'{self.current_driver_config.mode},'
               f'{self.current_driver_config.param1},'
               f'{self.current_driver_config.param2},'
               f'{self.current_driver_config.param3},'
               f'{self.current_driver_config.param4},'
               f'{self.current_driver_config.param5}')
        if wait_to_confirm:
            self.communicator.send_with_confirmation(msg, f'{self.communicator.DRIVER_CMD}:')
        else:
            self.communicator.send_and_forget(msg)

    def get_current_config(self):
        """Return the current configuration of this driver."""
        config = (f'{self.communicator.DRIVER_CMD}:{self.hw_number},{self.current_driver_config.trigger},'
                  f'{self.current_driver_config.switch_id},{self.current_driver_config.mode},'
                  f'{self.current_driver_config.param1},{self.current_driver_config.param2},'
                  f'{self.current_driver_config.param3},{self.current_driver_config.param4},'
                  f'{self.current_driver_config.param5}')

        self.log.debug("Current config for driver %s: %s", self.number, config)
        return config

    def get_board_name(self):
        """Return the board of this driver."""
        coil_index = 0
        for board_obj in self.communicator.platform.io_boards.values():
            if coil_index <= self.number < coil_index + board_obj.driver_count:
                return f"FAST Board {str(board_obj.node_id)}"
            coil_index += board_obj.driver_count

        # fall back if not found
        return "FAST Unknown Board"

    def get_recycle_ms_for_cmd(self, recycle, pulse_ms):
        """Return recycle ms."""
        if not recycle:
            return "00"
        if self.platform_settings.get('recycle_ms') is not None:
            return Util.int_to_hex_string(self.platform_settings['recycle_ms'])

        # default recycle_ms to pulse_ms * 2
        if pulse_ms * 2 > 255:
            return "FF"

        rms = Util.int_to_hex_string(pulse_ms * 2)
        self.log.debug("Using recycle_ms of %s for driver %s", rms, self.number)
        return rms

    async def reset(self):
        """Reset the driver to fresh (disabled/unconfigured) state."""
        self.log.debug("Resetting driver %s", self.number)
        self.current_driver_config.trigger = '00'
        self.current_driver_config.switch_id = '00'
        self.current_driver_config.mode = '00'
        self.current_driver_config.param1 = '00'
        self.current_driver_config.param2 = '00'
        self.current_driver_config.param3 = '00'
        self.current_driver_config.param4 = '00'
        self.current_driver_config.param5 = '00'

        await self.communicator.send_and_wait_for_response_processed(self.get_current_config(),
                                                                     self.get_current_config())

    def disable(self):
        """Disable (turn off) this driver."""
        if not self._reenable_autofire_if_configured():
            self.log.debug("Disabling driver %s", self.number)
            self.communicator.send_and_forget(f'{self.communicator.TRIGGER_CMD}:{self.hw_number},02')
        else:
            self.log.debug("Received disable command for driver %s but reenabled autofire instead", self.number)

    def set_hardware_rule(self, mode, switch, coil_settings, **kwargs):  # noqa: MC0001
        """Write a hardware rule to the FAST controller for this driver."""
        self.log.debug("Setting hardware rule for driver %s. Mode: %s. Switch: %s. coil_settings: %s. kwargs: %s.",
                       self.number, mode, switch.hw_switch.hw_number, coil_settings, kwargs)

        self._check_switch_coil_combination(switch, coil_settings.hw_driver)

        reconfigured = False
        trigger_needed = False
        switch_needed = False

        if not mode:
            mode = self.current_driver_config.mode

        new_settings = kwargs
        new_settings['pwm1_ms'] = Util.int_to_hex_string(coil_settings.pulse_settings.duration)
        new_settings['pwm1_power'] = Util.float_to_pwm8_hex_string(coil_settings.pulse_settings.power)

        if coil_settings.hold_settings:
            new_settings['pwm2_power'] = Util.float_to_pwm8_hex_string(coil_settings.hold_settings.power)

            if coil_settings.hold_settings.duration and coil_settings.hold_settings.duration <= 255:
                new_settings['pwm2_ms'] = Util.int_to_hex_string(coil_settings.hold_settings.duration,
                                                                 allow_overflow=True)
        else:
            # If there are no new hold settings, we want to use the existing ones since they could have been
            # configured via other commands which support more settings than these rules
            new_settings['pwm2_power'] = None
            new_settings['pwm2_ms'] = None

        if coil_settings.recycle is True:
            # MPF autofire rules will use True, so pull it from the config if specified.
            if self.platform_settings['recycle_ms']:
                new_settings['recycle_ms'] = Util.int_to_hex_string(self.platform_settings['recycle_ms'])
            else:
                new_settings['recycle_ms'] = '00'
        elif not coil_settings.recycle:  # False or None
            new_settings['recycle_ms'] = '00'
        else:
            new_settings['recycle_ms'] = Util.int_to_hex_string(coil_settings.recycle)

        # Update the current_driver_config with any new settings
        for idx, param in enumerate(self.mode_param_mapping[mode]):
            if param is None or new_settings.get(param, None) is None:
                continue
            if param == 'delay_ms_x10':
                new_settings['delay_ms'] = Util.int_to_hex_string(int(new_settings[param], 16) // 10)
            elif param == 'pwm2_ms_x100':
                new_settings['pwm2_ms'] = Util.int_to_hex_string(int(new_settings[param], 16) // 100)

            param_name = f'param{idx+1}'
            if new_settings[param] != getattr(self.current_driver_config, param_name):
                setattr(self.current_driver_config, param_name, new_settings[param])
                reconfigured = True

        if self.current_driver_config.switch_id != switch.hw_switch.hw_number:
            self.current_driver_config.switch_id = switch.hw_switch.hw_number
            trigger_needed = True
            switch_needed = True

        if self.current_driver_config.mode != mode:
            self.current_driver_config.mode = mode
            reconfigured = True

        trigger = '01'

        if switch.invert:
            trigger = self.set_bit(trigger, 4)

        # If this driver mode has an off switch, we need to check to see if it's inverted
        # and then update the trigger to act on inverted too. Since the off switch object
        # does not exist here, we need to go find it.
        if new_settings.get('off_switch', None):
            off_switch = self.communicator.switches[int(new_settings['off_switch'], 16)]
            if off_switch.baseline_switch_config.mode == '02':  # inverted
                trigger = self.set_bit(trigger, 5)

        if self.is_new_config_needed(self.current_driver_config.trigger, trigger):
            reconfigured = True
        elif trigger != self.current_driver_config.trigger:
            trigger_needed = True

        self.current_driver_config.trigger = trigger
        self.autofire_config = copy(self.current_driver_config)

        if reconfigured:  # Send a new driver config
            self.send_config_to_driver(one_shot=False)
        elif trigger_needed:  # We only need to update the triggers
            # Set the driver to automatic using the existing configuration
            if switch_needed:
                self.communicator.send_and_forget(
                    f'{self.communicator.TRIGGER_CMD}:{self.hw_number},00,{switch.hw_switch.hw_number}')
            else:
                self.communicator.send_and_forget(
                    f'{self.communicator.TRIGGER_CMD}:{self.hw_number},00')

    def _check_switch_coil_combination(self, switch, coil):
        # TODO move this to the communicator or something? Since it's only Nano?

        # V2 hardware can write rules across node boards
        if self.communicator.config['controller'] != 'nano':
            return

        # first 8 switches always work
        if 0 <= switch.hw_switch.number <= 7:
            return

        if self.get_board_name() != switch.hw_switch.get_board_name():
            raise AssertionError(f"Driver {coil.number} and switch {switch.hw_switch.number} "
                                 "are on different boards. Cannot apply rule!")

    def is_new_config_needed(self, current, new):
        """Determines if bits other than 6 and 7 changed.

        If so, it means we need a full new DL command not just TL update.
        """
        current_num = int(current, 16)
        new_num = int(new, 16)
        bitmask = 0x3F  # 0011 1111 in binary

        # Use bitwise AND to mask out bits 6 and 7
        current_num &= bitmask
        new_num &= bitmask

        return current_num != new_num

    def clear_autofire(self):
        """Clear autofire."""
        self.log.debug("Clearing autofire mode for driver %s", self.number)
        self._check_and_clear_delay()
        if not self.autofire_config:
            return

        # TL control code 2 sets bit 7 and clears 6
        self.current_driver_config.trigger = self.clear_bit(self.current_driver_config.trigger, 6)
        self.current_driver_config.trigger = self.set_bit(self.current_driver_config.trigger, 7)

        self.autofire_config = None
        self.communicator.send_and_forget(f'{self.communicator.TRIGGER_CMD}:{self.hw_number},02')

    def set_relay(self, relay_switch, debounce_closed_ms, debounce_open_ms):
        """Set an AC Relay rule with virtual switch."""
        self.log.debug("Setting A/C Relay for driver %s (0x%s) and switch %s (0x%s)",
                       self.number, self.hw_number, relay_switch.number, relay_switch.hw_number)
        self.current_driver_config = FastDriverConfig(
            number=self.hw_number, trigger='81',
            switch_id=relay_switch.hw_number,
            mode='25',
            param1=Util.int_to_hex_string(debounce_closed_ms),
            param2=Util.int_to_hex_string(debounce_open_ms),
            param3='00',
            param4='00',
            param5='00'
        )
        self.send_config_to_driver(wait_to_confirm=True)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable (turn on) this driver."""
        self.log.debug("Enabling (turning on) driver %s (0x%s) mode %s with pulse_settings: %s and hold_settings: %s.",
                       self.number, self.hw_number, self.current_driver_config.mode, pulse_settings, hold_settings)

        self._check_and_clear_delay()

        reconfigured = False
        mode = self.current_driver_config.mode

        # AC Relays have special behavior
        if mode == '25':
            self.log.debug(" - A/C Relay activating!")
            self.communicator.send_and_forget(f'{self.communicator.TRIGGER_CMD}:{self.hw_number},03')
            return

        pwm1_ms = Util.int_to_hex_string(pulse_settings.duration)
        pwm1_power = Util.float_to_pwm8_hex_string(pulse_settings.power)
        pwm2_power = Util.float_to_pwm8_hex_string(hold_settings.power)

        if mode != '18':
            mode = '18'
            reconfigured = True

        if self.current_driver_config.param1 != pwm1_ms:
            self.current_driver_config.param1 = pwm1_ms
            reconfigured = True

        if self.current_driver_config.param2 != pwm1_power:
            self.current_driver_config.param2 = pwm1_power
            reconfigured = True

        if self.current_driver_config.param3 != pwm2_power:
            self.current_driver_config.param3 = pwm2_power
            reconfigured = True

        if self.current_driver_config.param4 != '00':
            self.current_driver_config.param4 = '00'
            reconfigured = True

        if self.current_driver_config.mode != '18':
            self.current_driver_config.mode = '18'
            reconfigured = True

        if not reconfigured:
            # Trigger the driver directly using the existing configuration
            self.communicator.send_and_forget(f'{self.communicator.TRIGGER_CMD}:{self.hw_number},03')
            return
        # Send a new driver config and also trigger it now
        self.current_driver_config.trigger = 'C1'
        self.send_config_to_driver(one_shot=False)

    def timed_enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Pulse and hold this driver for a specified duration."""
        self.log.debug("Timed enabling (pulsing and holding) driver %s with pulse_settings: %s and hold_settings: %s.",
                       self.number, pulse_settings, hold_settings)

        if not hold_settings.duration and self.current_driver_config.mode == '70':
            # If we are in mode 70, timed enable defaults are already set
            hold_settings = None

        self._pulse(pulse_settings, hold_settings)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse this driver."""
        self.log.debug("Pulsing driver %s with pulse_settings: %s.",
                       self.number, pulse_settings)
        self._pulse(pulse_settings)

    def _pulse(self, pulse_settings: PulseSettings, hold_settings: HoldSettings = None):
        """Pulse this driver, with an optional hold setting.

        The FAST platform supports pulse and hold configuration in the same command, so
        this method can be used for both pulse() and timed_enable() behavior.
        """
        reconfigured = False
        mode = self.current_driver_config.mode

        # This method current hardcodes the params but some modes use different orders
        # The workaround for now is to just set the mode to 10.
        # No real downside, it will just send a full DL command instead of a TL
        # TODO elegant future implementation would be to use the mode_param_mapping to use the current configured mode

        if mode not in ['10', '12', '18', '70']:
            mode = '10'
            reconfigured = True

        pwm1_ms = Util.int_to_hex_string(pulse_settings.duration)
        pwm1_power = Util.float_to_pwm8_hex_string(pulse_settings.power)

        if self.current_driver_config.param1 != pwm1_ms:
            self.current_driver_config.param1 = pwm1_ms
            reconfigured = True

        if self.current_driver_config.param2 != pwm1_power:
            self.current_driver_config.param2 = pwm1_power
            reconfigured = True

        if hold_settings is not None:
            if hold_settings.duration > 25500:
                raise AssertionError("FAST platform does not support hold durations > 25500ms")
            if 25500 >= hold_settings.duration > 255:
                hold_ms = Util.int_to_hex_string(hold_settings.duration // 100)
                mode = '70'
            elif 255 >= hold_settings.duration >= 0:
                hold_ms = Util.int_to_hex_string(hold_settings.duration)
                mode = '10'

            hold_power = Util.float_to_pwm8_hex_string(hold_settings.power)

        else:
            hold_ms = self.current_driver_config.param3
            hold_power = self.current_driver_config.param4

        if self.current_driver_config.mode != mode:
            self.current_driver_config.mode = mode
            reconfigured = True

        if self.current_driver_config.param3 != hold_ms:
            self.current_driver_config.param3 = hold_ms
            reconfigured = True

        if self.current_driver_config.param4 != hold_power:
            self.current_driver_config.param4 = hold_power
            reconfigured = True

        if not reconfigured:
            # Trigger the driver directly using the existing configuration
            self.communicator.send_and_forget(f'{self.communicator.TRIGGER_CMD}:{self.hw_number},01')
        else:  # Send a new driver config and also trigger it now
            self.send_config_to_driver(one_shot=True)

        # reset not add so we keep on kicking this delay down the road if we keep getting pulses
        self.communicator.machine.delay.reset(pulse_settings.duration + 1, self._reenable_autofire_if_configured,
                                              f'fast_driver_{self.number}_delay')

    def _reenable_autofire_if_configured(self):
        """Reenable autofire if configured.

        Returns True if autofire was reenabled, False otherwise.
        """
        self._check_and_clear_delay()
        if self.autofire_config and self.autofire_config != self.current_driver_config:
            self.log.debug("Re-enabling autofire mode")
            self.current_driver_config = copy(self.autofire_config)
            self.send_config_to_driver()
            return True
        return False

    def _check_and_clear_delay(self):
        """Check if we have a delay and clear it."""
        self.communicator.machine.delay.remove(f'fast_driver_{self.number}_delay')


"""Text transitions used for segment displays."""
import abc
from typing import Optional, List

from mpf.core.placeholder_manager import TextTemplate
from mpf.core.rgb_color import RGBColor
from mpf.devices.segment_display.segment_display_text import SegmentDisplayText, UncoloredSegmentDisplayText

STEP_OUT_OF_RANGE_ERROR = "Step is out of range"
TRANSITION_DIRECTION_UNKNOWN_ERROR = "Transition uses an unknown direction value"


class TransitionBase(metaclass=abc.ABCMeta):

    """Base class for text transitions in segment displays."""

    __slots__ = ["output_length", "config", "collapse_dots", "collapse_commas", "use_dots_for_commas"]

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool,
                 use_dots_for_commas: bool, config: dict) -> None:
        """Initialize the transition."""
        self.output_length = output_length
        self.config = config
        self.collapse_dots = collapse_dots
        self.collapse_commas = collapse_commas
        self.use_dots_for_commas = use_dots_for_commas

        for key, value in config.items():
            if hasattr(self, key):
                setattr(self, key, value)

    @abc.abstractmethod
    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        raise NotImplementedError

    # pylint: disable=too-many-arguments
    @abc.abstractmethod
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        raise NotImplementedError


class TransitionRunner:

    """Class to run/execute transitions using an iterator."""

    __slots__ = ["_transition", "_step", "_current_placeholder", "_new_placeholder", "_current_colors", "_new_colors"]

    # pylint: disable=too-many-arguments
    def __init__(self, machine, transition: TransitionBase, current_text: str, new_text: str,
                 current_colors: Optional[List[RGBColor]] = None,
                 new_colors: Optional[List[RGBColor]] = None) -> None:
        """Class initializer."""
        self._transition = transition
        self._step = 0
        self._current_placeholder = TextTemplate(machine, current_text)
        self._new_placeholder = TextTemplate(machine, new_text)
        self._current_colors = current_colors
        self._new_colors = new_colors

    def __iter__(self):
        """Return the iterator."""
        return self

    def __next__(self):
        """Evaluate and return the next transition step."""
        if self._step >= self._transition.get_step_count():
            raise StopIteration

        transition_step = self._transition.get_transition_step(self._step,
                                                               self._current_placeholder.evaluate({}),
                                                               self._new_placeholder.evaluate({}),
                                                               self._current_colors,
                                                               self._new_colors)
        self._step += 1
        return transition_step


class NoTransition(TransitionBase):

    """Segment display no transition effect."""

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return 1

    # pylint: disable=too-many-arguments
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        return SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots, self.collapse_commas,
                                           self.use_dots_for_commas, new_colors)


class PushTransition(TransitionBase):

    """Segment display push transition effect."""

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                 config: dict) -> None:
        """Class initializer."""
        self.direction = 'right'
        self.text = None
        self.text_color = None
        super().__init__(output_length, collapse_dots, collapse_commas, use_dots_for_commas, config)

        if self.text is None:
            self.text = ''

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return self.output_length + len(self.text)

    # pylint: disable=too-many-arguments
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        current_display_text = SegmentDisplayText.from_str(current_text, self.output_length, self.collapse_dots,
                                                           self.collapse_commas, self.use_dots_for_commas,
                                                           current_colors)
        new_display_text = SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots,
                                                       self.collapse_commas, self.use_dots_for_commas, new_colors)

        if self.text:
            if new_colors and not self.text_color:
                text_color = [new_colors[0]]
            else:
                text_color = self.text_color
            transition_text = SegmentDisplayText.from_str(self.text, len(self.text), self.collapse_dots,
                                                          self.collapse_commas, self.use_dots_for_commas, text_color)
        else:
            transition_text = UncoloredSegmentDisplayText([], self.collapse_dots, self.collapse_commas,
                                                          self.use_dots_for_commas)

        if self.direction == 'right':
            temp_list = new_display_text
            temp_list.extend(transition_text)
            temp_list.extend(current_display_text)
            return temp_list[
                self.output_length + len(self.text) - (step + 1):2 * self.output_length + len(
                    self.text) - (step + 1)]

        if self.direction == 'left':
            temp_list = current_display_text
            temp_list.extend(transition_text)
            temp_list.extend(new_display_text)
            return temp_list[step + 1:step + 1 + self.output_length]

        raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)


class CoverTransition(TransitionBase):

    """Segment display cover transition effect."""

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                 config: dict) -> None:
        """Class initializer."""
        self.direction = 'right'
        self.text = None
        self.text_color = None
        super().__init__(output_length, collapse_dots, collapse_commas, use_dots_for_commas, config)

        if self.text is None:
            self.text = ''

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return self.output_length + len(self.text)

    # pylint: disable=too-many-arguments
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        current_display_text = SegmentDisplayText.from_str(current_text, self.output_length, self.collapse_dots,
                                                           self.collapse_commas, self.use_dots_for_commas,
                                                           current_colors)
        new_display_text = SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots,
                                                       self.collapse_commas, self.use_dots_for_commas, new_colors)

        if self.text:
            if new_colors and not self.text_color:
                text_color = [new_colors[0]]
            else:
                text_color = self.text_color
            transition_text = SegmentDisplayText.from_str(self.text, len(self.text), self.collapse_dots,
                                                          self.collapse_commas, self.use_dots_for_commas, text_color)
        else:
            transition_text = UncoloredSegmentDisplayText([], self.collapse_dots, self.collapse_commas,
                                                          self.use_dots_for_commas,)

        if self.direction == 'right':
            new_extended_display_text = new_display_text
            new_extended_display_text.extend(transition_text)

            if step < self.output_length:
                temp_text = new_extended_display_text[-(step + 1):]
                temp_text.extend(current_display_text[step + 1:])
            else:
                temp_text = new_display_text[-(step + 1):-(step + 1) + self.output_length]

            return temp_text

        if self.direction == 'left':
            new_extended_display_text = transition_text
            new_extended_display_text.extend(new_display_text)

            if step < self.output_length:
                temp_text = current_display_text[:self.output_length - (step + 1)]
                temp_text.extend(new_extended_display_text[:step + 1])
            else:
                temp_text = new_extended_display_text[step - self.output_length + 1:step + 1]

            return temp_text

        raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)


class UncoverTransition(TransitionBase):

    """Segment display uncover transition effect."""

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                 config: dict) -> None:
        """Class initializer."""
        self.direction = 'right'
        self.text = None
        self.text_color = None
        super().__init__(output_length, collapse_dots, collapse_commas, use_dots_for_commas, config)

        if self.text is None:
            self.text = ''

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return self.output_length + len(self.text)

    # pylint: disable=too-many-arguments
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        current_display_text = SegmentDisplayText.from_str(current_text, self.output_length, self.collapse_dots,
                                                           self.collapse_commas, self.use_dots_for_commas,
                                                           current_colors)
        new_display_text = SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots,
                                                       self.collapse_commas, self.use_dots_for_commas, new_colors)

        if self.text:
            if new_colors and not self.text_color:
                text_color = [new_colors[0]]
            else:
                text_color = self.text_color
            transition_text = SegmentDisplayText.from_str(self.text, len(self.text), self.collapse_dots,
                                                          self.collapse_commas, self.use_dots_for_commas, text_color)
        else:
            transition_text = UncoloredSegmentDisplayText([], self.collapse_dots, self.collapse_commas,
                                                          self.use_dots_for_commas)

        if self.direction == 'right':
            current_extended_display_text = transition_text
            current_extended_display_text.extend(current_display_text)

            if step < len(self.text):
                temp_text = current_extended_display_text[
                    len(self.text) - step - 1:len(self.text) - step - 1 + self.output_length]
            else:
                temp_text = new_display_text[:step - len(self.text) + 1]
                temp_text.extend(current_extended_display_text[:self.output_length - len(temp_text)])

            return temp_text

        if self.direction == 'left':
            current_extended_display_text = current_display_text
            current_extended_display_text.extend(transition_text)

            if step < len(self.text):
                temp_text = current_extended_display_text[step + 1:step + 1 + self.output_length]
            else:
                temp_text = current_display_text[step + 1:]
                temp_text.extend(new_display_text[-(self.output_length - len(temp_text)):])
            return temp_text

        raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)


class WipeTransition(TransitionBase):

    """Segment display wipe transition effect."""

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                 config: dict) -> None:
        """Class initializer."""
        self.direction = 'right'
        self.text = None
        self.text_color = None
        super().__init__(output_length, collapse_dots, collapse_commas, use_dots_for_commas, config)

        if self.text is None:
            self.text = ''

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return self.output_length + len(self.text)

    # pylint: disable=too-many-arguments,too-many-branches,too-many-return-statements
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        current_display_text = SegmentDisplayText.from_str(current_text, self.output_length, self.collapse_dots,
                                                           self.collapse_commas, self.use_dots_for_commas,
                                                           current_colors)
        new_display_text = SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots,
                                                       self.collapse_commas, self.use_dots_for_commas, new_colors)

        if self.text:
            if new_colors and not self.text_color:
                text_color = [new_colors[0]]
            else:
                text_color = self.text_color
            transition_text = SegmentDisplayText.from_str(self.text, len(self.text), self.collapse_dots,
                                                          self.collapse_commas, self.use_dots_for_commas, text_color)
        else:
            transition_text = UncoloredSegmentDisplayText([], self.collapse_dots, self.collapse_commas,
                                                          self.use_dots_for_commas)

        if self.direction == 'right':
            if step < len(self.text):
                temp_text = transition_text[-(step + 1):]
                temp_text.extend(current_display_text[step + 1:])
            elif step < self.output_length:
                temp_text = new_display_text[:step - len(self.text) + 1]
                temp_text.extend(transition_text)
                temp_text.extend(current_display_text[len(temp_text):])
            else:
                temp_text = new_display_text[:step - len(self.text) + 1]
                temp_text.extend(transition_text[:self.output_length - len(temp_text)])
            return temp_text

        if self.direction == 'left':
            if step < len(self.text):
                temp_text = current_display_text[:self.output_length - (step + 1)]
                temp_text.extend(transition_text[:step + 1])
            elif step < self.output_length:
                temp_text = current_display_text[:self.output_length - (step + 1)]
                temp_text.extend(transition_text)
                temp_text.extend(new_display_text[len(temp_text):])
            elif step < self.output_length + len(self.text) - 1:
                temp_text = transition_text[step - (self.output_length + len(self.text)) + 1:]
                temp_text.extend(new_display_text[-(self.output_length - len(temp_text)):])
            else:
                temp_text = new_display_text
            return temp_text

        raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)


class SplitTransition(TransitionBase):

    """Segment display split transition effect."""

    # pylint: disable=too-many-arguments
    def __init__(self, output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                 config: dict) -> None:
        """Class initializer."""
        self.direction = 'out'
        self.mode = 'push'
        super().__init__(output_length, collapse_dots, collapse_commas, use_dots_for_commas, config)

    def get_step_count(self):
        """Return the total number of steps required for the transition."""
        return int((self.output_length + 1) / 2)

    # pylint: disable=too-many-arguments,too-many-branches,too-many-return-statements
    def get_transition_step(self, step: int, current_text: str, new_text: str,
                            current_colors: Optional[List[RGBColor]] = None,
                            new_colors: Optional[List[RGBColor]] = None) -> SegmentDisplayText:
        """Calculate all the steps in the transition."""
        if step < 0 or step >= self.get_step_count():
            raise AssertionError(STEP_OUT_OF_RANGE_ERROR)

        current_display_text = SegmentDisplayText.from_str(current_text, self.output_length, self.collapse_dots,
                                                           self.collapse_commas, self.use_dots_for_commas,
                                                           current_colors)
        new_display_text = SegmentDisplayText.from_str(new_text, self.output_length, self.collapse_dots,
                                                       self.collapse_commas, self.use_dots_for_commas, new_colors)

        if self.mode == 'push':
            if self.direction == 'out':
                if step == self.get_step_count() - 1:
                    return new_display_text

                characters = int(self.output_length / 2)
                split_point = characters
                if characters * 2 == self.output_length:
                    characters -= 1
                else:
                    split_point += 1

                characters -= step
                temp_text = current_display_text[split_point - characters:split_point]
                temp_text.extend(new_display_text[characters:characters + (self.output_length - 2 * characters)])
                temp_text.extend(current_display_text[split_point:split_point + characters])
                return temp_text

            if self.direction == 'in':
                if step == self.get_step_count() - 1:
                    return new_display_text

                split_point = int(self.output_length / 2)
                characters = 1
                if split_point * 2 < self.output_length:
                    split_point += 1

                characters += step
                temp_text = new_display_text[split_point - characters:split_point]
                temp_text.extend(current_display_text[characters:characters + (self.output_length - 2 * characters)])
                temp_text.extend(new_display_text[split_point:split_point + characters])
                return temp_text

            raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)

        if self.mode == 'wipe':
            if self.direction == 'out':
                if step == self.get_step_count() - 1:
                    return new_display_text

                characters = int(self.output_length / 2)
                if characters * 2 == self.output_length:
                    characters -= 1

                characters -= step
                temp_text = current_display_text[:characters]
                temp_text.extend(new_display_text[characters:characters + (self.output_length - 2 * characters)])
                temp_text.extend(current_display_text[-characters:])
                return temp_text

            if self.direction == 'in':
                if step == self.get_step_count() - 1:
                    return new_display_text

                temp_text = new_display_text[:step + 1]
                temp_text.extend(current_display_text[step + 1:step + 1 + (self.output_length - 2 * len(temp_text))])
                temp_text.extend(new_display_text[-(step + 1):])
                return temp_text

            raise AssertionError(TRANSITION_DIRECTION_UNKNOWN_ERROR)

        raise AssertionError("Transition uses an unknown mode value")



"""FAST Neuron Serial Communicator."""
# mpf/platforms/fast/communicators/net_neuron.py

from packaging import version

from mpf.core.utility_functions import Util
from mpf.platforms.fast.communicators.base import FastSerialCommunicator
from mpf.platforms.fast.fast_io_board import FastIoBoard
from mpf.exceptions.config_file_error import ConfigFileError
from mpf.platforms.fast.fast_driver import FastDriverConfig
from mpf.platforms.fast.fast_switch import FASTSwitch
from mpf.platforms.fast.fast_driver import FASTDriver
from mpf.platforms.fast import fast_defines


class FastNetNeuronCommunicator(FastSerialCommunicator):

    """Serial communicator for the Neuron board."""

    MIN_FW = version.parse('2.06')
    IO_MIN_FW = version.parse('1.09')
    MAX_IO_BOARDS = 9
    MAX_SWITCHES = 104
    MAX_DRIVERS = 48
    IGNORED_MESSAGES = ['WD:P', 'TL:P']
    TRIGGER_CMD = 'TL'
    DRIVER_CMD = 'DL'
    SWITCH_CMD = 'SL'

    __slots__ = ["io_loop", "switches", "drivers"]

    def __init__(self, platform, processor, config):
        """Initialize the Neuron controller."""
        super().__init__(platform, processor, config)

        self.io_loop = [None] * len(self.config['io_loop'])
        self.switches = list()
        self.drivers = list()

        self.message_processors['SA:'] = self._process_sa
        self.message_processors['CH:'] = self._process_ch
        self.message_processors['!B:'] = self._process_boot_message
        self.message_processors['\x11\x11!'] = self._process_reboot_done
        self.message_processors['NN:'] = self._process_nn
        self.message_processors[f'{self.DRIVER_CMD}:'] = self.process_driver_config_msg
        self.message_processors[f'{self.SWITCH_CMD}:'] = self.process_switch_config_msg
        self.message_processors[f'/{self.SWITCH_CMD[-1]}:'] = self._process_switch_open
        self.message_processors[f'-{self.SWITCH_CMD[-1]}:'] = self._process_switch_closed

        for board, board_config in self.config['io_loop'].items():
            board_config['index'] = int(board_config['order']) - 1

            try:
                self.io_loop[board_config['index']] = board
            except IndexError:
                raise ConfigFileError(f"Invalid order value for I/O board '{board}'. "
                                      "Order values must be sequential starting at 1.", 7, self.log.name)  # TODO

    async def init(self):
        """Initialize the hardware switches and drivers."""
        self.create_switches()
        self.create_drivers()
        # Loop here until we get a response
        await self.send_and_wait_for_response_processed('ID:', 'ID:', max_retries=-1)
        await self.configure_hardware()
        # Force expire the watchdog since who knows what state the board is in?
        self.send_and_forget('WD:1')
        await self.query_io_boards()

    async def configure_hardware(self):
        """Send the hardware configuration command."""
        await self.send_and_wait_for_response_processed(
            f'CH:{fast_defines.HARDWARE_KEY[self.platform.machine_type]},FF', 'CH:')

    def create_switches(self):
        """Track switches for the FAST loop.

        Neuron tracks all switches regardless of how many are connected.
        """
        for i in range(self.MAX_SWITCHES):
            self.switches.append(FASTSwitch(self, i))

    def create_drivers(self):
        """Track drivers for the FAST loop.

        Neuron tracks all drivers regardless of how many are connected.
        """
        for i in range(self.MAX_DRIVERS):
            self.drivers.append(FASTDriver(self, i))

    async def soft_reset(self, **kwargs):
        """Reset the NET processor."""
        del kwargs

        await self.reset_switches()
        await self.reset_drivers()

        # We call the platform version of this meth since it uses the wait event
        await self.platform.get_hw_switch_states(query_hw=True)

    async def clear_board_serial_buffer(self):
        """Clear out the serial buffer."""
        # Not needed on the Neuron

    def _process_boot_message(self, msg):
        """Reset the NET CPU."""
        del msg
        self.log.info("Resetting NET Processor...")
        # TODO Do something?

    def _process_reboot_done(self, msg):
        del msg
        self.log.info("Processor reset complete.")
        # TODO Mark all configs as dirty

    async def reset_switches(self):
        """Query the NET processor to get a list of switches and their configurations.

        Compare that to how they should be configured, and send new configuration
        commands for any that are different.
        """
        # all physical switches, not just ones defined in the config
        for switch in self.switches:
            await self.send_and_wait_for_response_processed(
                f'{self.SWITCH_CMD}:{switch.hw_number}', f'{self.SWITCH_CMD}:{switch.hw_number}')

        self.platform.switches_initialized = True

    async def reset_drivers(self):
        """Query the NET processor to get a list of drivers and their configurations.

        Compare that to how they should be configured, and send new configuration
        commands for any that are different.
        """
        # self.drivers contains a list of all drivers, not just ones defined in the config
        for driver in self.drivers:
            await self.send_and_wait_for_response_processed(
                f'{self.DRIVER_CMD}:{Util.int_to_hex_string(driver.number)}', f"{self.DRIVER_CMD}:")

        self.platform.drivers_initialized = True

    def _process_ch(self, msg):
        if msg == 'F':
            raise AssertionError("Could not configure hardware. Check your FAST config.")
        self.done_processing_msg_response()

    async def query_io_boards(self):
        """Walk through the I/O boards in the config to verify the physical boards match the config."""
        if 'switches' not in self.machine.config and 'coils' not in self.machine.config:
            self.log.debug("No coils or switches configured. Skipping I/O board discovery.")
            return

        current_node = 0
        while current_node < len(self.config['io_loop']):
            await self.send_and_wait_for_response_processed('NN:{:02X}'.format(current_node), 'NN:')

            # Don't move on until we get board 00 in since it can take a sec after a reset
            if current_node + 1 >= len(self.platform.io_boards):
                current_node += 1

    def _process_nn(self, msg):
        firmware_ok = True

        if msg == 'F':  # NN:F
            return  # TODO now what?

        node_id, model, fw, dr, sw, _, _, _, _, _, _ = msg.split(',')

        node_id = Util.hex_string_to_int(node_id)
        dr = Util.hex_string_to_int(dr)
        sw = Util.hex_string_to_int(sw)
        model = model.strip('\x00')
        model = ('-').join(model.split('-')[:3])  # Remove the revision dash if it's there

        if not model or model == '!Node Not Found!':
            self.done_processing_msg_response()  # TODO good candidate for retry option
            return

        if node_id in self.platform.io_boards:
            self.done_processing_msg_response()
            return

        name = self.io_loop[node_id]
        # Fp-I/O-3208-2 -> FP-I/O-3208
        model_string_from_config = ('-').join(self.config['io_loop'][name]['model'].split('-')[:3]).upper()

        assert model == model_string_from_config, \
            f"I/O board config error. Board {node_id} is reporting as model {model}, " + \
            f"but the config file says it\'s model {model_string_from_config}"

        prior_sw = 0
        prior_drv = 0

        for i in range(node_id):
            prior_sw += self.platform.io_boards[i].switch_count
            prior_drv += self.platform.io_boards[i].driver_count

        self.platform.register_io_board(FastIoBoard(self, name, node_id, model, fw, sw, dr, prior_sw, prior_drv))

        self.log.info('Registered I/O Board %s: Model: %s, Firmware: %s, Switches: %s, Drivers: %s',
                      node_id, model, fw, sw, dr)

        min_fw = self.IO_MIN_FW  # TODO move to IO board class
        if min_fw > version.parse(fw):
            self.platform.log.critical("Firmware version mismatch. MPF requires the I/O boards "
                                       "to be firmware %s, but your Board %s (%s) is firmware %s",
                                       min_fw, node_id, model, fw)
            firmware_ok = False

        if not firmware_ok:
            raise AssertionError("Exiting due to I/O board firmware mismatch")

        self.done_processing_msg_response()

    def process_driver_config_msg(self, msg):
        """Handle an incoming DL/DN: message."""
        if msg == 'P':
            self.done_processing_msg_response()
            return

        # From here down we're processing driver config data, 9 fields, one byte each
        # <driver_id>,<trigger>,<switch_id>,<mode>,<param_1>,<param_2>,<param_3>,<param_4>,<param_5>
        # https://fastpinball.com/fast-serial-protocol/net/dl/

        current_hw_driver_config = FastDriverConfig(*msg.split(','))

        try:
            driver_obj = self.drivers[int(current_hw_driver_config.number, 16)]
        except IndexError:
            return  # we might get a config for a driver that doesn't exist in MPF, no worries

        if driver_obj.current_driver_config != current_hw_driver_config:
            driver_obj.send_config_to_driver()
            # We'll mark this done with we get the :P response back above
        else:
            self.done_processing_msg_response()

    def process_switch_config_msg(self, msg):
        """Incoming SL:<switch_id>,<mode>,<debounce_close>,<debounce_open> message."""
        if msg == 'P':  # SL:P
            self.done_processing_msg_response()
            return

        # From here down we're processing switch config data
        # '00,02,01,02' = switch number, mode, debounce close, debounce open
        # https://fastpinball.com/fast-serial-protocol/net/sl/

        switch, mode, debounce_close, debounce_open = msg.split(',')

        try:
            switch_obj = self.switches[int(switch, 16)]
        except IndexError:  # Neuron tracks 104 switches regardless of how many are connected
            return

        if (switch_obj.current_hw_config.mode != mode or
                switch_obj.current_hw_config.debounce_close != debounce_close or
                switch_obj.current_hw_config.debounce_open != debounce_open):
            # TODO this seems tedious, should we switch the dataclass to a namedtuple?
            switch_obj.send_config_to_switch()
            # it will get marked done via the :P response above
        else:
            self.done_processing_msg_response()

    async def update_switches_from_hardware(self):
        """Process an SA: message and update switch states."""
        await self.send_and_wait_for_response_processed('SA:', 'SA:')

    def _process_sa(self, msg):
        if not self.platform.switches_initialized:
            # This means the switches have not been initialized and this data is garbage
            return

        hw_states = {}
        _, raw_switch_data = msg.split(',')

        for offset, byte in enumerate(bytearray.fromhex(raw_switch_data)):
            for i in range(8):

                num = (offset * 8) + i

                if byte & (2**i):
                    hw_states[num] = 1
                else:
                    hw_states[num] = 0

        self.platform.hw_switch_data = hw_states
        self.update_switches_from_hw_data()
        self.done_processing_msg_response()

    def update_switches_from_hw_data(self):
        """Uses the existing platform.hw_switch_data dict to update switch states in the machine's switch controller.

        This will silently sync the switch.hw_state. If the logical state changes,
        it will process it like any switch change.
        """
        for switch in self.machine.switches:
            hw_state = self.platform.hw_switch_data[switch.hw_switch.number]

            if hw_state != switch.hw_state:
                switch.hw_state = hw_state

            logical_state = switch.invert ^ hw_state

            if logical_state != switch.state:
                self.machine.switch_controller.process_switch_obj(switch, logical_state, True)

        self.platform.new_switch_data.set()  # Signal that we have new switch data

    def _process_switch_open(self, msg):
        """Process local switch open.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        self.machine.switch_controller.process_switch_by_num(state=0,
                                                             num=int(msg, 16),
                                                             platform=self.platform,
                                                             logical=True)

    def _process_switch_closed(self, msg):
        """Process local switch closed.

        Args:
        ----
            msg: switch number
            remote_processor: Processor which sent the message.
        """
        self.machine.switch_controller.process_switch_by_num(state=1,
                                                             num=int(msg, 16),
                                                             platform=self.platform,
                                                             logical=True)

    def stopping(self):
        """Stop the Neuron processor and disable the watchdog."""
        self.send_and_forget('WD:1')



"""FAST Pinball hardware platform."""

import asyncio
from typing import Dict, Optional

from serial import SerialException

from mpf.core.platform import (RgbDmdPlatform, DriverConfig, DriverSettings,
                               LightsPlatform, RepulseSettings,
                               SegmentDisplayPlatform, ServoPlatform,
                               SwitchConfig, SwitchSettings)
from mpf.core.utility_functions import Util
from mpf.exceptions.config_file_error import ConfigFileError
from mpf.exceptions.runtime_error import MpfRuntimeError
from mpf.platforms.fast import fast_defines
from mpf.platforms.fast.fast_audio import FASTAudioInterface
from mpf.platforms.fast.fast_dmd import FASTDMD
from mpf.platforms.fast.fast_driver import FASTDriver
from mpf.platforms.fast.fast_gi import FASTGIString
from mpf.platforms.fast.fast_io_board import FastIoBoard
from mpf.platforms.fast.fast_led import (FASTRGBLED, FASTLEDChannel,
                                         FASTExpLED)
from mpf.platforms.fast.fast_light import FASTMatrixLight
from mpf.platforms.fast.fast_port_detector import FastPortDetector
from mpf.platforms.fast.fast_segment_display import FASTSegmentDisplay
from mpf.platforms.fast.fast_servo import FastServo
from mpf.platforms.fast.fast_switch import FASTSwitch
# pylint: disable-msg=too-many-instance-attributes
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.platforms.system11 import System11OverlayPlatform


class FastHardwarePlatform(ServoPlatform, LightsPlatform, RgbDmdPlatform,
                           SegmentDisplayPlatform,
                           System11OverlayPlatform):

    """Platform class for the FAST Pinball hardware."""

    __slots__ = ["config", "configured_ports", "machine_type", "is_retro",
                 "serial_connections", "fast_rgb_leds", "fast_exp_leds", "fast_segs",
                 "exp_boards_by_address", "exp_boards_by_name", "exp_breakout_boards",
                 "exp_breakouts_with_leds", "hw_switch_data", "new_switch_data",
                 "io_boards", "io_boards_by_name", "switches_initialized",
                 "drivers_initialized", "audio_interface"]

    port_types = ['net', 'exp', 'aud', 'dmd', 'rgb', 'seg', 'emu']

    def __init__(self, machine):
        """Initialize FAST hardware platform.

        Args:
        ----
            machine: The main ``MachineController`` instance.
        """
        super().__init__(machine)

        self.config = self.machine.config_validator.validate_config("fast", self.machine.config['fast'])
        self._configure_device_logging_and_debug("FAST", self.config, url_base='https://fastpinball.com/mpf/error')

        self.configured_ports = list()

        for port_type in self.port_types:
            if self.config[port_type]:
                self.configured_ports.append(port_type)

        try:
            self.machine_type = self.config["net"]["controller"]
        except KeyError:
            self.machine_type = 'no_net'

        if self.machine_type in ['sys11', 'wpc89', 'wpc95']:
            self.debug_log("Configuring the FAST Controller for Retro driver board")
            self.is_retro = True
        elif self.machine_type in ['neuron', 'nano']:
            self.debug_log("Configuring FAST Controller for FAST I/O boards.")
            self.is_retro = False
        elif self.machine_type == 'no_net':
            pass
        else:
            self.raise_config_error(f'Unknown machine_type "{self.machine_type}" configured fast.', 6)

        # Even though System11 uses ticks, that's handled on the Overlay and not needed here.
        self.features['tickless'] = True
        # Most FAST platforms don't use ticks, but System11 does
        #self.features['tickless'] = self.machine_type != 'sys11'
        self.features['max_pulse'] = 25500

        self.serial_connections = dict()
        self.fast_rgb_leds = dict()
        self.fast_exp_leds = dict()
        self.fast_segs = list()
        self.exp_boards_by_address = dict()  # k: EE address, v: FastExpansionBoard instances
        self.exp_boards_by_name = dict()  # k: str name, v: FastExpansionBoard instances
        self.exp_breakout_boards = dict()  # k: EEB address, v: FastBreakoutBoard instances
        self.exp_breakouts_with_leds = set()
        self.hw_switch_data = {i: 0 for i in range(112)}
        self.new_switch_data = asyncio.Event()  # Used to signal when we have updated switch data
        self.io_boards = dict()     # type: Dict[int, FastIoBoard]  # TODO move to NET communicator(s) classes?
        self.io_boards_by_name = dict()     # type: Dict[str, FastIoBoard]
        self.switches_initialized = False
        self.drivers_initialized = False
        self.audio_interface = None

    def get_info_string(self):
        """Dump info strings about attached FAST hardware."""
        info_string = ""

        for port in sorted(self.serial_connections.keys()):
            info_string += f"{port.upper()}: {self.serial_connections[port].remote_model} " + \
                f"v{self.serial_connections[port].remote_firmware}\n"

        info_string += "\nI/O Boards:\n"
        for board in self.io_boards.values():
            info_string += board.get_description_string() + "\n"
        return info_string

    async def initialize(self):
        """Initialize platform."""
        # self.machine.events.add_async_handler('machine_reset_phase_1', self.soft_reset)
        self.machine.events.add_async_handler('init_phase_1', self.soft_reset)
        self.machine.events.add_handler('init_phase_3', self._start_communicator_tasks)
        self.machine.events.add_handler('machine_reset_phase_2', self._init_complete)
        await self._connect_to_hardware()

    async def soft_reset(self, **kwargs):
        """Soft reset the FAST controller.

        Used to reset / sync / verify all hardware configurations. This command does not perform a
        hard reset of the boards, rather it queries the boards for their current configuration and
        reapplies (with warnings) any configs that are out of sync.

        This command runs during the init_phase_1 phase, and then skips the reset phases, then runs
        again on subsequent resets during the machine_reset_phase_1 phase.
        """
        del kwargs
        self.debug_log("Soft resetting FAST platform.")

        for comm in self.serial_connections.values():
            await comm.soft_reset()

    def _init_complete(self, **kwargs):
        del kwargs
        # Runs on init_phase_2 and moves soft_reset() from init_phase_1 to machine_reset_phase_1
        # We need the soft_reset() to run earlier on boot, but then at reset from there on out
        self.machine.events.remove_handler(self.soft_reset)
        self.machine.events.add_async_handler('machine_reset_phase_1', self.soft_reset)

    def stop(self):
        """Stop platform and close connections."""
        # TODO move all this into the comm classes
        if not self.unit_test:  # Only do this with real hardware TODO better way to check?
            for conn in self.serial_connections.values():
                # clear out whatever's in the send queues

                # TODO register a diverter callback which just swallows all messages
                # Then set the diverter and clear the queue

                for _ in range(conn.send_queue.qsize()):
                    conn.send_queue.get_nowait()
                    conn.send_queue.task_done()
                # conn.msg_diverter.set()

        for conn in self.serial_connections.values():
            conn.stopping()
            conn.cancel_tasks()

        # wait 100ms for the messages to be sent
        if not self.unit_test:
            self.machine.clock.loop.run_until_complete(asyncio.sleep(.1))

        for port, connection in self.serial_connections.items():
            if connection:
                connection.stop()
                self.serial_connections[port] = None

        self.serial_connections = dict()

    def _start_communicator_tasks(self, **kwargs):  # init_phase_3
        del kwargs
        for comm in self.serial_connections.values():
            comm.start_watchdog()
            comm.start_tasks()

    def __repr__(self):
        """Return str representation."""
        return '[FAST Platform Interface]'

    def register_io_board(self, board):  # TODO move to NET communicator(s) classes?
        """Register a FAST I/O Board.

        Args:
        ----
            board: 'mpf.platform.fast.fast_io_board.FastIoBoard' to register
        """
        if board.node_id in self.io_boards:
            raise AssertionError("Duplicate node_id")
        self.io_boards[board.node_id] = board
        self.io_boards_by_name[board.name] = board

    def register_expansion_board(self, board):  # TODO move to EXP communicator(s) classes?
        """Register an Expansion board."""
        if board.address in self.exp_boards_by_address:
            raise AssertionError("Duplicate expansion board address")
        self.exp_boards_by_address[board.address] = board
        self.exp_boards_by_name[board.name] = board

    def register_breakout_board(self, board):  # TODO move to EXP communicator(s) classes?
        """Register a Breakout board."""
        if board.address in self.exp_breakout_boards:
            raise AssertionError("Duplicate breakout board address")
        self.exp_breakout_boards[board.address] = board

    def register_led_board(self, board):  # TODO move to EXP communicator(s) classes?
        """Register a Breakout board that has LEDs."""
        self.exp_breakouts_with_leds.add(board.address[:3])

    async def _connect_to_hardware(self):  # TODO move to class methods?
        """Connect to each port from the config."""
        await self._check_for_autodetect()

        for port in self.configured_ports:

            config = self.config[port]

            # pylint: disable=import-outside-toplevel
            if port == 'net':
                if config['controller'] == 'neuron':
                    from mpf.platforms.fast.communicators.net_neuron import \
                        FastNetNeuronCommunicator
                    communicator = FastNetNeuronCommunicator(platform=self, processor=port, config=config)
                elif config['controller'] == 'nano':
                    from mpf.platforms.fast.communicators.net_nano import \
                        FastNetNanoCommunicator
                    communicator = FastNetNanoCommunicator(platform=self, processor=port, config=config)
                elif config['controller'] in ['sys11', 'wpc89', 'wpc95']:
                    from mpf.platforms.fast.communicators.net_retro import \
                        FastNetRetroCommunicator
                    communicator = FastNetRetroCommunicator(platform=self, processor=port, config=config)
                else:
                    raise AssertionError("Unknown controller type")  # TODO better error
                self.serial_connections['net'] = communicator

            elif port == 'exp':
                from mpf.platforms.fast.communicators.exp import \
                    FastExpCommunicator
                communicator = FastExpCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['exp'] = communicator
            elif port == 'seg':
                from mpf.platforms.fast.communicators.seg import \
                    FastSegCommunicator
                communicator = FastSegCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['seg'] = communicator
            elif port == 'aud':
                from mpf.platforms.fast.communicators.aud import \
                    FastAudCommunicator
                communicator = FastAudCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['aud'] = communicator
            elif port == 'dmd':
                from mpf.platforms.fast.communicators.dmd import \
                    FastRgbDmdCommunicator
                communicator = FastRgbDmdCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['dmd'] = communicator
            elif port == 'emu':
                from mpf.platforms.fast.communicators.emu import \
                    FastEmuCommunicator
                communicator = FastEmuCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['emu'] = communicator
            elif port == 'rgb':
                from mpf.platforms.fast.communicators.rgb import \
                    FastRgbCommunicator
                communicator = FastRgbCommunicator(platform=self, processor=port, config=config)
                self.serial_connections['rgb'] = communicator
            else:
                raise AssertionError("Unknown processor type")  # TODO better error

            try:
                await communicator.connect()
            except SerialException as e:
                if config.get("optional"):
                    self.info_log("Unable to connect to %s on port %s, flagged as optional so ignoring",
                                  port, config['port'])
                    del self.serial_connections[port]
                    continue
                raise MpfRuntimeError(f"Could not open serial port {port}. Is something else connected to the port? "
                                      "Did the port number or your computer change? Do you have permissions "
                                      "to access the port?", 1, self.log.name) from e
            await communicator.init()
            self.serial_connections[port] = communicator

    async def _check_for_autodetect(self):
        # Figures out which processors need to be autodetected and then runs the autodetect process
        autodetect_processors = list()
        hardcoded_ports = list()

        for processor in self.port_types:
            try:
                for port in self.config[processor]['port']:
                    if port == 'auto':
                        autodetect_processors.append(processor)
                    else:
                        hardcoded_ports.append(port)
            except KeyError:
                pass

        if not autodetect_processors:
            return

        detector = FastPortDetector(platform=self,
                                    autodetect_processors=autodetect_processors,
                                    hardcoded_ports=hardcoded_ports)

        await detector.detect_ports()

    async def get_hw_switch_states(self, query_hw=False):
        """Return a dict of hw switch states.

        If query_hw is True, will re-query the hardware for the current switch states. Otherwise it will just return
        the last cached value.
        """
        # If the switches have not been initialized then their states are garbage, so don't bother
        if self.switches_initialized and query_hw:
            self.new_switch_data.clear()
            await self.serial_connections['net'].update_switches_from_hardware()
            await self.new_switch_data.wait()  # Wait here until we have updated switch data

        return self.hw_switch_data

    def _parse_driver_number(self, number):
        # Accepts FAST driver number string (is3208-1) and returns the driver index (0-based int)

        try:
            board_str, driver_str = number.split("-")

        except ValueError:  # If there's no dash, assume it's a driver number
            return int(number, 16)

        try:
            board = self.io_boards_by_name[board_str]
        except KeyError:
            raise AssertionError(f"I/O Board {board_str} does not exist for driver {number}")

        driver = int(driver_str)

        if board.driver_count <= driver:
            raise AssertionError(f"I/O Board {board} only has drivers 0-{board.driver_count-1}. "
                                 f"Driver value {driver} is not valid.")

        index = board.start_driver + driver

        if index + 1 > self.serial_connections['net'].MAX_DRIVERS:
            raise AssertionError(
                f"I/O Board {board} driver {driver} is out of range. This would be driver {index + 1} "
                f"but this platform supports a max of {self.serial_connections['net'].MAX_DRIVERS} drivers.")

        return index

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> FASTDriver:
        """Configure a driver.

        Parameters
        ----------
            config: Driver config.
            number: string number entry from config (e.g. 'io3208-0)
            platform_settings: Platform specific settings.

        Returns
        -------
        Driver object
        """
        if not self.serial_connections['net']:
            raise AssertionError('A request was made to configure a FAST '
                                 'driver, but no connection to a NET processor'
                                 'is available')

        if not number:
            raise AssertionError("Driver needs a number")

        # For a Retro Controller, look up the driver number
        if self.is_retro:
            try:
                index = int(fast_defines.RETRO_DRIVER_MAP[number.upper()], 16)
            except KeyError:
                self.raise_config_error(f"Could not find Retro driver {number}", 1)

        # If we have FAST I/O boards, parse the config into a FAST hex driver number
        elif self.machine_type in ['nano', 'neuron']:
            index = self._parse_driver_number(number)

        else:
            raise AssertionError("Invalid machine type: {self.machine_type}")

        driver = self.serial_connections['net'].drivers[index]  # contains all drivers on the board
        # platform.drivers is empty at this point
        driver.set_initial_config(config, platform_settings)

        return driver

    async def configure_servo(self, number: str, config: dict) -> FastServo:
        """Configure a servo.

        Args:
        ----
            number: Number of servo
            config: Dict of config settings.

        Returns: Servo object.
        """
        # TODO consolidate with similar code in configure_light()
        number = number.lower()
        parts = number.split("-")

        exp_board = self.exp_boards_by_name[parts[0]]

        try:
            _, port = parts
            breakout_id = '0'
        except ValueError:
            _, breakout_id, port = parts
            breakout_id = breakout_id.strip('b')

        brk_board = exp_board.breakouts[breakout_id]

        # verify this board support servos
        assert int(port) <= int(brk_board.features['servo_ports'])  # TODO should this be stored as an int?

        config.update(self.machine.config_validator.validate_config('fast_servos', config['platform_settings']))
        del config['platform_settings']

        return FastServo(brk_board, port, config)

    def _parse_switch_number(self, number):
        try:
            board_str, switch_str = number.split("-")
        except ValueError as e:
            total_switches = 0
            for board_obj in self.io_boards.values():
                total_switches += board_obj.switch_count
            index = Util.int_to_hex_string(number)

            if int(index, 16) >= total_switches:
                raise AssertionError(f"Switch {int(index, 16)} does not exist. Only "
                                     f"{total_switches} switches found. Switch number: {number}") from e

            return index

        try:
            board = self.io_boards_by_name[board_str]
        except KeyError:
            raise AssertionError(f"Board {board_str} does not exist for switch {number}")

        switch = int(switch_str)

        if board.switch_count <= switch:
            raise AssertionError(f"Board {board} only has switches 0-{board.switch_count-1}. "
                                 f"Switch value {switch} is not valid.")

        return Util.int_to_hex_string(board.start_switch + switch)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> FASTSwitch:
        """Configure the switch object for a FAST Pinball controller.

        V1 FAST Controllers support two types of switches: `local` and `network`.
        Local switches are switches that are connected to the FAST controller
        board itself, and network switches are those connected to a FAST I/O
        board. V2 FAST Controllers consider all switches to be local, even those
        connected to I/O boards.

        MPF needs to know which type of switch is this is. You can specify the
        switch's connection type in the config file via the ``connection:``
        setting (either ``local`` or ``network``).

        If a connection type is not specified, this method will use some
        intelligence to try to figure out which default should be used.

        If the DriverBoard type is ``fast`` and the firmware is legacy (v1),
        then mpf assumes the default is ``network``. If it's a v2 firmware or
        any other type of board (``sys11``, ``wpc95``, ``wpc89``) then mpf
        assumes the connection type is ``local``. Connection types can be mixed
        and matched in the same machine.

        Args:
        ----
            number: Number of this switch.
            config: Switch config.
            platform_config: Platform specific settings.

        Returns: Switch object.
        """
        if not number:
            raise AssertionError("Switch needs a number")

        if not self.serial_connections['net']:
            raise AssertionError("A request was made to configure a FAST "
                                 "switch, but no connection to a NET processor"
                                 "is available")

        if self.is_retro:
            # translate switch num to FAST switch
            try:
                number = fast_defines.RETRO_SWITCH_MAP[str(number).upper()]
            except KeyError:
                self.raise_config_error(f"Could not find switch {number}", 2)
        else:
            try:
                number = self._parse_switch_number(number)
            except ValueError:
                self.raise_config_error(f"Could not parse switch number {config.name}/{number}. Seems "
                                        "to be not a valid switch number for the FAST platform.", 8)

        switch = self.serial_connections['net'].switches[int(number, 16)]
        switch.set_initial_config(config, platform_config)

        return switch

    # pylint: disable-msg=too-many-locals
    def configure_light(self, number, subtype, config, platform_settings) -> LightPlatformInterface:
        """Configure light in platform."""
        del platform_settings

        if subtype == "gi":
            return FASTGIString(number, self.serial_connections['net'], self.machine,
                                int(1 / self.config['net']['gi_hz'] * 1000))
        if subtype == "matrix":
            return FASTMatrixLight(number, self.serial_connections['net'], self.machine,
                                   int(1 / self.config['net']['lamp_hz'] * 1000), self)
        if not subtype or subtype == "led":
            # make everything lowercase and strip trailing channel number
            parts, channel = number.lower().rsplit('-', 1)
            # split into board name, breakout, port, led
            parts = parts.split('-')

            if parts[0] in self.exp_boards_by_name:
                # this is an expansion board LED in config file format
                exp_board = self.exp_boards_by_name[parts[0]]

                try:
                    _, port, led = parts
                    breakout = '0'
                except ValueError:
                    _, breakout, port, led = parts
                    breakout = breakout.strip('b')

                # ports are always 1-4, but some EXP boards have more which are labeled 5-8
                # Those are really 1-4 of the next breakout board, so if we get a port > 4
                # then sort it out to the real internal values
                if int(port) > 4:
                    # assume 4 LED ports per breakout, could change to a lookup
                    breakout = str((int(port) - 1) // 4)
                    port = str((int(port) - 1) % 4 + 1)

                try:
                    brk_board = exp_board.breakouts[breakout]
                except KeyError:
                    # TODO change to mpf config exception
                    raise AssertionError(f'Board {exp_board} does not have a config entry for Breakout {breakout}')

                index = self.port_idx_to_hex(port, led, 32, config.name)
                this_led_number = f'{brk_board.address}{index}'

                # this code runs once for each channel, so it will be called 3x per LED which
                # is why we check this here
                if this_led_number not in self.fast_exp_leds:
                    self.fast_exp_leds[this_led_number] = FASTExpLED(this_led_number,
                                                                     exp_board.config['led_fade_time'], self)

                fast_led_channel = FASTLEDChannel(self.fast_exp_leds[this_led_number], channel)
                self.fast_exp_leds[this_led_number].add_channel(int(channel), fast_led_channel)

            elif int(parts[0]) > 255:
                # EXP LED in int form, which is how "previous:" values are calculated

                raw_hex_string = hex(int(parts[0]))[2:]  # lowercase with 0x prefix stripped"
                this_led_number = Util.normalize_hex_string(raw_hex_string, len(raw_hex_string))

                exp_board = self.exp_boards_by_address[this_led_number[:2]]

                if this_led_number not in self.fast_exp_leds:
                    # RGBW LEDs could span multiple FAST LEDs, so make sure it exists
                    self.fast_exp_leds[this_led_number] = FASTExpLED(this_led_number,
                                                                     exp_board.config['led_fade_time'], self)

                fast_led_channel = FASTLEDChannel(self.fast_exp_leds[this_led_number], channel)
                self.fast_exp_leds[this_led_number].add_channel(int(channel), fast_led_channel)

            else:
                # Nano LED

                try:
                    number = self.port_idx_to_hex(parts[0], parts[1], 64)
                except IndexError:
                    # this is a legacy LED number as an int
                    number = f'{int(parts[0]):02X}'

                if number not in self.fast_rgb_leds:
                    try:
                        self.fast_rgb_leds[number] = FASTRGBLED(number, self)
                    except KeyError:
                        # This number is not valid
                        raise ConfigFileError(f"Invalid LED number: {'_'.join(parts)}", 3, self.log.name)

                fast_led_channel = FASTLEDChannel(self.fast_rgb_leds[number], channel)
                self.fast_rgb_leds[number].add_channel(int(channel), fast_led_channel)

            return fast_led_channel
        raise AssertionError(f"Unknown light subtype {subtype}")

    def port_idx_to_hex(self, port, device_num, devices_per_port, name=None):
        """Converts port number and LED index into the proper FAST hex number.

        port: the LED port number printed on the board. First port is 1. No zeros.
        device_num: LED position in the change, First LED is 1. No zeros.
        devices_per_port: number of LEDs per port. Typically 32.
        name: used for config error logging

        Returns: FAST hex string for the LED
        """
        port = int(port)
        device_num = int(device_num)

        if device_num < 1:
            raise AssertionError(f"Device number {device_num} is not valid for device {name}. "
                                 "The first device in the change should be 1, not 0")

        if port < 1:
            raise AssertionError(f"Port {port} is not valid for device {device_num}")

        if device_num > devices_per_port:
            if name:
                # TODO get a final error code
                self.raise_config_error(f"Device number {device_num} exceeds the number of devices per port "
                                        f"({devices_per_port}) for LED {name}", 8)
            else:
                raise AssertionError(f"Device number {device_num} exceeds the number of devices per port "
                                     f"({devices_per_port})")

        port_offset = (port - 1) * devices_per_port
        device_num = device_num - 1
        return f'{(port_offset + device_num):02X}'

    def parse_light_number_to_channels(self, number: str, subtype: str = "led"):
        """Transform an MPF config light number to a FAST channel.

        Takes the `number:` and `subtype:` from the config file and parses and
        standardizes it into a format the FAST Light interface can understand

        Incoming `number:` examples:

        number: playfield-b2-2-3 --> board: playfield, breakout: 2, port: 2, led: 3
        number: playfield-b2-2-3-0 --> board: playfield, breakout: 2, port: 2, led: 3, channel: 0
        number: playfield-2-3 --> board: playfield, (breakout:1, not listed), port: 2, led: 3
        number: playfield-2-3-0 --> board: playfield, (breakout:1, not listed), port: 2, led: 3, channel: 0
        number: 15 --> Nano LED 15
        number: 15-0 --> Nano LED 15, channel 0
        """
        if not subtype or subtype == "led":
            return self._parse_led_light_number(number)
        if subtype == "gi":
            return self._parse_gi_light_number(number)
        if subtype == "matrix":
            return self._parse_matrix_light_number(number)

        raise AssertionError(f"Unknown LED subtype {subtype}")

    def _parse_led_light_number(self, number):
        parts = number.lower().split("-")

        if parts[0] in self.exp_boards_by_name:
            # This is an expansion board LED
            if not parts[1].startswith('b'):
                # No breakout specified, so we insert a b0
                parts.insert(1, 'b0')

            if len(parts) == 4:
                # No channel specified, so we return 3 channels 0,1,2
                return [{'number': '-'.join(parts) + f'-{i}'} for i in range(3)]

            if len(parts) == 5:
                # We have a channel specified
                channel = int(parts[4])
                if 0 <= channel <= 2:
                    result = []
                    for i in range(3):
                        working_parts = parts.copy()
                        if i + channel > 2:
                            # Channel rolls over, increment the LED number
                            working_parts[3] = str(int(working_parts[3]) + 1)
                            working_parts[4] = str((channel + i) % 3)
                        else:
                            working_parts[4] = str(channel + i)
                        result.append({'number': '-'.join(working_parts)})
                    return result
                raise AssertionError(f"Invalid LED channel: {channel}")
            raise AssertionError(f"Invalid LED number: {number}")

        # This is a Nano LED
        if '-' in str(number):
            # num = list(map(int, str(number).split('-')))
            # index = num[0] * 64 + num[1]
            index = int(self.port_idx_to_hex(parts[0], parts[1], 64), 16)
        else:
            index = int(number)

        return [{"number": f"{index}-{i}"} for i in range(3)]

    def _parse_gi_light_number(self, number):
        if self.is_retro:  # translate matrix/map number to FAST GI number
            try:
                number = fast_defines.RETRO_GI_MAP[str(number).upper()]
            except KeyError:
                self.raise_config_error(f"Could not find GI {number}", 3)
        else:
            number = Util.int_to_hex_string(number)
        return [{"number": number}]

    def _parse_matrix_light_number(self, number):
        if self.is_retro:  # translate matrix number to FAST light num
            try:
                number = fast_defines.RETRO_LIGHT_MAP[str(number).upper()]
            except KeyError:
                self.raise_config_error(f"Could not find light {number}", 4)
        else:
            number = Util.int_to_hex_string(number)
        return [{"number": number}]

    def configure_rgb_dmd(self, name):
        """Configure a hardware DMD connected to a FAST controller."""
        if not self.serial_connections['dmd']:
            raise AssertionError("A request was made to configure a FAST DMD, "
                                 "but no connection to a DMD processor is "
                                 "available.")

        return FASTDMD(self.machine, name, self.serial_connections['dmd'])

    def configure_audio_interface(self):
        """Configure a hardware FAST audio controller."""
        if 'aud' not in self.serial_connections:
            self.log.debug("Skipping FAST Audio Interface because there's no 'aud:' section in the FAST config.")
            return None

        return FASTAudioInterface(self.machine, self.serial_connections['aud'])

    async def configure_segment_display(self, number: str, display_size: int, platform_settings) -> FASTSegmentDisplay:
        """Configure a segment display."""
        self.debug_log("Configuring FAST segment display.")
        del platform_settings
        if not self.serial_connections['seg']:
            raise AssertionError("A request was made to configure a FAST "
                                 "Segment Display but no connection is "
                                 "available.")

        return FASTSegmentDisplay(int(number), self.serial_connections['seg'])

    @classmethod
    def get_coil_config_section(cls):
        """Return coil config section."""
        return "fast_coils"

    @classmethod
    def get_switch_config_section(cls):
        """Return switch config section."""
        return "fast_switches"

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse continues. Typically used for
        autofire coils such as pop bumpers.

        FAST Driver Mode 10 or 70, depending on settings
        """
        coil.hw_driver.set_hardware_rule(None, enable_switch, coil)
        # TODO currently this will just use whatever the current mode is. Should we do some math and force a mode?

    def set_delayed_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings, delay_ms: int):
        """Set pulse on hit and release rule to driver.

        When a switch is hit and a certain delay passed it pulses a driver.
        When the switch is released the pulse continues.
        Typically used for kickbacks.

        FAST Driver Mode 30
        """
        coil.hw_driver.set_hardware_rule('30', enable_switch, coil, delay_ms=delay_ms)

    def set_pulse_on_hit_and_release_rule(self, enable_switch, coil):
        """Set pulse on hit and release rule to driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse is canceled. Typically used on
        the main coil for dual coil flippers without eos switch.

        FAST Driver Mode 18 (with pwm2_power = 00)
        """
        # Force hold to None which is needed with this rule
        coil.hold_settings = None
        coil.hw_driver.set_hardware_rule('18', enable_switch, coil)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and release rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver gets disabled. Typically used for single coil flippers.

        FAST Driver Mode 18 (with pwm2_power != 00)
        """
        # coil.hw_driver.set_pulse_on_hit_and_enable_and_release_rule(enable_switch, coil)
        coil.hw_driver.set_hardware_rule('18', enable_switch, coil)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. When the switch is released
        the pulse is canceled and the driver gets disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes disabled. Typically used on the main coil for dual-wound coil flippers with eos switch.

        FAST Driver Mode 75
        """
        del repulse_settings  # TODO do we want to implement software repulse?
        # If enabled, set a switch rule to look for EOS being open and flipper button closed and manually pulse?
        off_switch = eos_switch.hw_switch.hw_number

        coil.hw_driver.set_hardware_rule('75', enable_switch, coil, off_switch=off_switch)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver becomes disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes enabled (likely with PWM).
        Typically used on the coil for single-wound coil flippers with eos switch.

        FAST Driver Mode 20
        """
        coil.hw_driver.set_hardware_rule('20', enable_switch, coil, eos_switch=eos_switch,
                                         repulse_settings=repulse_settings)

    def clear_hw_rule(self, switch, coil):
        """Clear a hardware rule.

        This is used if you want to remove the linkage between a switch and
        some driver activity. For example, if you wanted to disable your
        flippers (so that a player pushing the flipper buttons wouldn't cause
        the flippers to flip), you'd call this method with your flipper button
        as the *sw_num*.

        Args:
        ----
            switch: The switch whose rule you want to clear.
            coil: The coil whose rule you want to clear.

        """
        self.debug_log("Clearing HW Rule for switch: %s, coils: %s",
                       switch.hw_switch.number, coil.hw_driver.number)

        # TODO: check that the rule is switch + coil and not another switch + this coil

        driver = coil.hw_driver

        driver.clear_autofire()



"""Segment Display serial communicator interface."""
# mpf/platforms/fast/communicators/seg.py

from packaging import version

from mpf.platforms.fast.communicators.base import FastSerialCommunicator

MIN_FW = version.parse('0.01')         # Minimum FW for a Segment Display

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class FastSegCommunicator(FastSerialCommunicator):

    """Handles the serial communication to the FAST platform."""

    IGNORED_MESSAGES = []

    async def init(self):
        """Ping the display board for an ID."""
        # Loop here until we get a response
        await self.send_and_wait_for_response_processed('ID:', 'ID:', max_retries=-1)

    def start_tasks(self):
        """Setup the interval tasks for updating the display."""
        for s in self.machine.device_manager.collections["segment_displays"]:
            self.platform.fast_segs.append(s.hw_display)

        self.platform.fast_segs.sort(key=lambda x: x.number)

        if self.platform.fast_segs:
            self.tasks.append(self.machine.clock.schedule_interval(self._update_segs,
                              1 / self.config['fps']))

    def _update_segs(self, **kwargs):
        del kwargs
        for s in self.platform.fast_segs:

            if s.next_text:
                self.send_and_forget(f'PA:{s.hex_id},{s.next_text.convert_to_str()[0:7]}')
                s.next_text = None

            if s.next_color:
                self.send_and_forget(('PC:{},{}').format(s.hex_id, s.next_color))
                s.current_color = s.next_color
                s.next_color = None

    async def soft_reset(self):
        """Trigger a soft reset of all segments."""
        # TODO turn off all segments

    def stopping(self):
        """Stop the segment display and clear all text/colors."""
        for s in self.platform.fast_segs:
            self.send_and_forget(f'PA:{s.hex_id},')
            s.next_text = None
            s.next_color = None
            s.current_color = None



"""Contains the base classes for spinners."""

from mpf.core.enable_disable_mixin import EnableDisableMixinSystemWideDevice

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.system_wide_device import SystemWideDevice
from mpf.exceptions.config_file_error import ConfigFileError


MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor(_active="active", _idle="idle")
class Spinner(EnableDisableMixinSystemWideDevice, SystemWideDevice):

    """Represents a spinner or spinner group in a pinball machine."""

    config_section = 'spinners'
    collection = 'spinners'
    class_label = 'spinner'

    __slots__ = ["hits", "_active_ms", "_active", "_idle", "_event_buffer_ms", "delay"]

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialize spinner device."""
        super().__init__(machine, name)
        self._active = False
        self._idle = True
        self._active_ms = None
        self._event_buffer_ms = None
        self.hits = None
        self.delay = DelayManager(machine)
        self.enabled = True  # Default to enabled

        self.configure_logging(f'Spinner.{name}')

    async def _initialize(self):
        await super()._initialize()
        self.hits = 0
        # Cache this value because it's used a lot in rapid succession
        self._active_ms = self.config['active_ms']
        if self.config['max_events_per_second'] > 0:
            self._event_buffer_ms = (1 / self.config['max_events_per_second']) * 1000
            self.log.debug("Configured event buffer for %s", self._event_buffer_ms)
        # Can't read the switch until the switch controller is set up
        self.machine.events.add_handler('init_phase_4',
                                        self._register_switch_handlers, priority=1)

    def _register_switch_handlers(self, **kwargs):
        del kwargs

        labels = dict(zip(self.config['switches'], self.config['labels'])) if self.config['labels'] else None
        for switch in self.config['switches']:
            callback_kwargs = {"label": labels[switch]} if labels else None
            # register for notification of switch active state
            self.machine.switch_controller.add_switch_handler_obj(
                switch,
                self._update_state_from_switch, 1, callback_kwargs=callback_kwargs)

    def _update_state_from_switch(self, **kwargs):
        if not self.enabled:
            return
        label = kwargs.get("label")
        if not self._active:
            self.machine.events.post("spinner_{}_active".format(self.name), label=label)
            '''event: spinner_(name)_active
            desc: The idle spinner (name) was just hit and became active.

            This event will post whenever a spinner switch is hit and the spinner
            is not already active.

            args:
            label: The label of the switch that triggered the activation
            '''
            if label:
                self.machine.events.post("spinner_{}_{}_active".format(self.name, label))
                '''event: spinner_(name)_(label)_active
                desc: The idle spinner (name) was just hit and became active.

                This event will post whenever a spinner switch is hit and the spinner
                is not already active, but only if labels are defined for the spinner.
                '''
            self._active = True
            self._idle = False
        self.hits += 1

        if not self._event_buffer_ms or not self.delay.check("event_buffer"):
            self._post_hit_event(label=label, last_hits=self.hits - 1)

    def _post_hit_event(self, **kwargs):
        last_hits = kwargs.get("last_hits", 0)
        self.log.debug("Buffer check has %s previous hits, current is %s", last_hits, self.hits)
        if last_hits and last_hits == self.hits:
            self.delay.remove("event_buffer")
            return

        label = kwargs.get("label")

        self.machine.events.post("spinner_{}_hit".format(self.name), hits=self.hits,
                                 change=self.hits - last_hits, label=label)
        '''event: spinner_(name)_hit
        desc: The spinner (name) was just hit.

        This event will post whenever a spinner switch is hit.

        args:
        hits: The number of switch hits the spinner has had since it became active
        label: The label of the switch that was hit
        '''
        if label:
            self.machine.events.post("spinner_{}_{}_hit".format(self.name, label),
                                     hits=self.hits, change=self.hits - last_hits)
            '''event: spinner_(name)_(label)_hit
            desc: The spinner (name) was just hit on the switch labelled (label).

            This event will post whenever a spinner switch is hit and labels
            are defined for the spinner
            '''
        self.delay.reset(self._active_ms, self._deactivate, "deactivate")

        if self._event_buffer_ms:
            self.delay.add(self._event_buffer_ms, self._post_hit_event, "event_buffer",
                           label=label, last_hits=self.hits)

    def _deactivate(self, **kwargs):
        """Post an 'inactive' event after no switch hits for the active_ms duration."""
        del kwargs
        self.machine.events.post("spinner_{}_inactive".format(self.name), hits=self.hits)
        '''event: spinner_(name)_inactive
        desc: The spinner (name) is no longer receiving hits

        This event will post whenever a spinner has not received hits and
        its active_ms has timed out.

        args:
        hits: The number of switch hits the spinner had while it was active
        '''
        self._active = False
        if self.config['idle_ms']:
            self.delay.reset(self.config['idle_ms'], self._on_idle, "idle")
            if self.config['reset_when_inactive']:
                self.hits = 0
        else:
            self._idle = True

    def _on_idle(self, **kwargs):
        """Post an 'idle' event if the spinner has been inactive for the idle_ms duration."""
        del kwargs
        self.machine.events.post("spinner_{}_idle".format(self.name), hits=self.hits)
        '''event: spinner_(name)_idle
        desc: The spinner (name) is now idle

        This event will post whenever a spinner has not received hits and
        its idle_ms has timed out. If no idle_ms is defined, this event
        will not post.

        args:
        hits: The number of switch hits the spinner had while it was active
        '''
        self.hits = 0
        self._idle = True

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None):
        """Validate and parse spinner config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        for switch in config['switch']:
            if switch not in config['switches']:
                config['switches'].append(switch)

        if config['labels'] and len(config['labels']) != len(config['switches']):
            raise ConfigFileError("Spinner labels must be the same number as switches", 1, self.name)

        return config

    @property
    def active(self):
        """Return whether the spinner is actively spinning."""
        return self._active

    @property
    def idle(self):
        """Return whether the spinner is idle."""
        return self._idle



"""Test the virtual segment display connector plugin."""
from unittest.mock import patch, call, ANY

from mpf.core.rgb_color import RGBColor
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.tests.MpfBcpTestCase import MpfBcpTestCase


class TestVirtualSegmentDisplayConnector(MpfBcpTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        self.machine_config_patches['bcp'] = \
            {"connections": {"local_display": {"type": "mpf.tests.test_BcpMc.TestBcpClient"}}}
        self.machine_config_patches['bcp']['servers'] = []

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/virtual_segment_display_connector/'

    def setUp(self):
        self.machine_config_patches['mpf']['plugins'] = [
            'mpf.plugins.virtual_segment_display_connector.VirtualSegmentDisplayConnector']
        super().setUp()

    def get_use_bcp(self):
        return True

    @patch("mpf.core.bcp.bcp_interface.BcpInterface.bcp_trigger_client")
    def test_plugin(self, mock_bcp_trigger_client):
        client = self.machine.bcp.transport.get_named_client("local_display")
        self.assertIsNotNone(client)
        self.machine.bcp.interface.add_registered_trigger_event_for_client(client, 'update_segment_display')
        self.advance_time_and_run()

        display1 = self.machine.segment_displays["display1"]
        display2 = self.machine.segment_displays["display2"]
        display3 = self.machine.segment_displays["display3"]  # Should not have virtual connector set

        self.assertIsNotNone(display1.virtual_connector)
        self.assertIsNotNone(display2.virtual_connector)
        self.assertIsNone(display3.virtual_connector)

        display1.add_text("NEW TEXT")
        display1.set_color([RGBColor("FF0000"), RGBColor("00FF00")])
        self.assertTrue(mock_bcp_trigger_client.called)
        mock_bcp_trigger_client.assert_has_calls([call(client=ANY, flashing='False', flash_mask='',
                                                       name='update_segment_display', segment_display_name='display1',
                                                       text='EW TEXT', colors=['ffffff', 'ffffff', 'ffffff',
                                                                               'ffffff', 'ffffff', 'ffffff', 'ffffff']),
                                                  call(client=ANY, name='update_segment_display',
                                                       segment_display_name='display1', text='EW TEXT',
                                                       flashing='False', flash_mask='',
                                                       colors=['ff0000', '00ff00','00ff00', '00ff00', '00ff00',
                                                               '00ff00', '00ff00'])])
        mock_bcp_trigger_client.reset_mock()

        display2.add_text_entry("OTHER TEXT", [RGBColor("green")], FlashingType.FLASH_ALL, "", None, None, None, None)
        self.assertTrue(mock_bcp_trigger_client.called)
        mock_bcp_trigger_client.assert_has_calls([call(client=ANY, flashing='True', flash_mask='',
                                                       name='update_segment_display', segment_display_name='display2',
                                                       text='ER TEXT',
                                                       colors=[RGBColor("green").hex] * 7)])
        mock_bcp_trigger_client.reset_mock()

        display2.set_flashing(FlashingType.FLASH_MATCH)
        self.assertTrue(mock_bcp_trigger_client.called)
        mock_bcp_trigger_client.assert_has_calls([call(client=ANY, flashing='match', flash_mask='',
                                                       name='update_segment_display', segment_display_name='display2',
                                                       text='ER TEXT',
                                                       colors=[RGBColor("green").hex] * 7)
                                                  ])
        mock_bcp_trigger_client.reset_mock()

        display2.set_flashing(FlashingType.FLASH_MASK, "______FFFF")
        self.assertTrue(mock_bcp_trigger_client.called)
        mock_bcp_trigger_client.assert_has_calls([call(client=ANY, flashing='mask', flash_mask='______FFFF',
                                                       name='update_segment_display', segment_display_name='display2',
                                                       text='ER TEXT',
                                                       colors=[RGBColor("green").hex] * 7)
                                                  ])
        mock_bcp_trigger_client.reset_mock()

        display3.add_text("IGNORED")
        self.assertFalse(mock_bcp_trigger_client.called)
